Module NetworkTransformationCheck.

Load "LTS.SyncConditions". (* LTS transformation network and related notions *)
Export SyncConditions.
Load "Utils.ClassicalExt".
Import ClassicalExt.

Definition RuleSystem_cascading {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) :=
  forall I J, let IJ := Union nat I J in
    Disjoint nat I J ->
      forall s p q (BI  : BB (SystemLTS (LeftNetwork RSk I))  (SystemLTS (RightNetwork RSk I)))
                   (BIJ : BB (SystemLTS (LeftNetwork RSk IJ)) (SystemLTS (RightNetwork RSk IJ))), 
        NetInit (LeftNetwork RS J) q -> (BI s p <-> BIJ (Cmps s q J) (Cmps p q J)).

(* All k-extended patterns are bisimilar *)
Definition PatternNetworks_bisimilar {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) := 
  forall I, exists (B : BB (SystemLTS (LeftNetwork RSk I)) (SystemLTS (RightNetwork RSk I))), True.

(* ==================================================================
     Construction of bismulation relation C
   ================================================================== *)

Section proposition_proof.
(* Variables that will used in all Lemmas and Propositions *)
Context {n} {M} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) (TrM : @NetTransform n M RS).

Lemma Cas_Cmps_state_prop : 
  forall (v : @SyncLaw n) (s sm phm : State n) (X : Ensemble nat), 
    let I := Intersection nat X (Ac v) in
    let J := Setminus nat X (Ac v) in
      (forall i, I i -> nth i sm dummy = nth i phm dummy) ->
      (forall i, i < n -> ~X i -> nthS i phm = d_state) ->
        (Cmps (Cmps (d_staten n) phm J) (Cmps (d_staten n) sm I) I) = phm.
Proof.
intros v s sm phm X I J Heq HSphm.
apply (lstn_eq _ _ dummy) ; intros i Hin.
destruct (classic (I i)) as [HIi | HIi].
  (* I i *)
  rewrite (Cmps_i_in_I _ _ I i Hin HIi) ; rewrite (Cmps_i_in_I _ _ I i Hin HIi).
  rewrite (Heq i HIi).
  reflexivity.
  (* ~I x *)
  rewrite (Cmps_i_not_in_I (Cmps (d_staten n) phm J) _ I i Hin HIi).
  destruct (classic (J i)) as [HJi | HJi].
    (* J x *)  rewrite (Cmps_i_in_I _ _ J i Hin HJi) ; reflexivity.
    (* ~J x *) rewrite (Cmps_i_not_in_I _ _ J i Hin HJi).
    rewrite d_staten_nth.
    unfold I in HIi ; unfold J in HJi.
    specialize HSphm with i.
    unfold nthS, d_state in HSphm .
    apply list1_lst_eq in HSphm ; try exact Hin.
    rewrite HSphm ; reflexivity.
    unfold not ; intro HMsi.
    destruct (classic (Ac v i)) as [HAci | HAci].
      (* Ac v x *)  apply HIi ; split ; assumption.
      (* ~Ac v x *) apply HJi ; split ; assumption.
Qed.

Definition Bhat_rel s p :=
    let I := Union nat (MatchLIds TrM s) (MatchRIds TrM p) in
    let SysL X := SystemLTS (LeftNetwork X I) in
    let SysR X := SystemLTS (RightNetwork X I) in
      exists sm pm (B : BB (SysL RSk) (SysR RSk)),
           (* sm maps to s, pm maps to p *)
           MapL TrM I sm s /\ MapR TrM I pm p 
           (* sm and pm are related via B_{M(s)} *)
        /\ BI_rel RSk I B sm pm.

Definition D_rel (i : nat)(s p : State n) :=
    States (nth i (Procs M) d_lts) (nthS i s) /\ nthS i s = nthS i p.

Definition C_rel (s p : State n) :=
  let I := Union nat (MatchLIds TrM s) (MatchRIds TrM p) in
    forall i, i < n ->   (~I i -> D_rel i s p)
                      /\ (I i -> Bhat_rel s p).

Definition C := C_rel. (* Shorthand notation for relation C *)

Lemma C_related_NetMatchIds_prop : forall s p, C s p -> MatchLIds TrM s = MatchRIds TrM p.
Proof.
intros.
unfold C, C_rel, D_rel, Bhat_rel in H.
apply Extensionality_Ensembles.
unfold Same_set, Included, In.
split ; intros i HMi ;
  pose proof HMi as [Hin _] ;
  specialize H with i ;
  intuition ;
  [ destruct H1 as [_ [q [_ [_ [[HS [HSq HMq]] _]]]]]
  | destruct H1 as [q [_ [_ [[HS [HSq HMq]] _]]]]
  ] ;
  try first [ apply Union_introl ; exact HMi | apply Union_intror ; exact HMi ] ;
  pose proof (m_states _ _ (mLs TrM i)) (nthS i q) (nthS i s) ;
  pose proof (m_states _ _ (mR _ (TrG TrM i)) (nthS i q) (nthS i p)) ;
  [ assert (mR (mLs TrM i) (TrG TrM i) (nthS i q) (nthS i p)) by 
      ( apply HMq ; try exact Hin ; apply Union_introl ; exact HMi )
  | assert (mLs TrM i (nthS i q) (nthS i s)) by 
      (apply HMq ; try exact Hin ; apply Union_intror ; exact HMi ) 
  ] ;
  split ; try exact Hin ;
  exists (nthS i q) ; intuition.
Qed.

Lemma C_related_NetMatchIds_i_prop : forall s p, C s p -> forall i, MatchLIds TrM s i = MatchRIds TrM p i.
Proof.
intros.
rewrite (C_related_NetMatchIds_prop s p H).
reflexivity.
Qed.

(* C relates initial states *)
Lemma C_rel_init : 
  PatternNetworks_bisimilar RSk -> 
    forall s, (Init (SystemLTS M) s -> exists p, Init (SystemLTS TrM) p /\ C s p)
           /\ (Init (SystemLTS TrM) s -> exists p, Init (SystemLTS M) p /\ C p s).
Proof.
simpl.
unfold NetInit.
intros HPatternNetworks_bisimilar s.
unfold C, C_rel, D_rel, Bhat_rel.
split ; intro ; exists s ;
  [ pose proof H as HIM ; assert (Init (SystemLTS TrM) s) as HITrM
  | pose proof H as HITrM ; assert (Init (SystemLTS M) s) as HIM 
  ] ;
  simpl ; unfold NetInit ;
  intros ; try (specialize H with i ; intuition) ;
  try (first
  [ rewrite (Procs_cond TrM i H0) ;
    rewrite (IT_cond (mLs TrM i) (TrG TrM i))
  | rewrite <- (IT_cond (mLs TrM i) (TrG TrM i)) ;
    rewrite <- (Procs_cond TrM i H0)
  ] ; try assumption) ;
assert (States (SystemLTS M) s) as HSM by (apply init_cond ; auto) ;
assert (States (SystemLTS TrM) s) as HSTrM by (apply init_cond ; auto) ;
simpl in HIM, HITrM ; unfold NetInit in HIM, HITrM ;
set (I := Union nat (MatchLIds TrM s) (MatchRIds TrM s)) ;
split ; try assumption ;
intros ; (split ; [intuition | ]) ;
destruct (HPatternNetworks_bisimilar I) as [BI _] ;
destruct (exists_mapLState TrM I s) as [p [HMp HIpd]] ; try exact HSM ;
  try ( intros x HIx ; unfold I in HIx ;
        destruct HIx as [x HIx | x HIx]; try exact HIx ;
        destruct HIx as [Hxn [p Hmp]] ;
        split ; try exact Hxn ;
        exists p ;
        pose proof (HITrM x Hxn) as HIs ;
        rewrite Procs_cond in HIs ; try exact Hxn ;
        pose proof (m_init _ _ (mR _ (TrG TrM x)) (nthS x s) HIs p Hmp) as HIp ;
        rewrite <- Init_L_to_Init_R in HIp ;
        rewrite (mR_init_cond _ (TrG TrM x)) ; [exact Hmp | exact HIp] ) ;
  pose proof (MapL_init TrM _ _ HIM p HMp) as HIp ;
  intros ;
  exists p, p, BI ;
  split ; try exact HMp ;
  split ; 
  (* BI p p *)
  try (apply BI_rel_glue_states_related ; try exact HIp) ;
  (* MapR p s *)
  destruct HMp as [_ [_ HMp]] ;
  split ; try assumption ;
  split ; first 
  [ (* p is a state in the right network *)
    rewrite Net_Init_L_to_Init_R in HIp ;
    apply init_cond ; exact HIp
  | (* p right-maps to s *)
    intros ; specialize HMp with i0 ;
    pose proof (m_init _ _ _ _ (HIM i0 H2) _ (HMp H2 H3)) ;
    rewrite <- mR_init_cond ; intuition ].
Qed.

(* C relates states of M to states of T(M) *)
Lemma C_rel_states : 
  forall s p, C s p -> States (SystemLTS M) s /\ States (SystemLTS TrM) p.
Proof.
intros.
unfold C, C_rel, D_rel, Bhat_rel in H.
simpl.
unfold NetStates.
set (I := Union nat (MatchLIds TrM s) (MatchRIds TrM p)).
split ; intros ;
specialize H with i ;
destruct (EnsembleNat_dec I i) as [HMsi | HNMsi] ; intuition ;
  (* i in I *)
  [ destruct H1 as [sm [pm [_ [[HSq _] _]]]]
  | destruct H1 as [sm [pm [_ [_ [[HSq _] _]]]]]
  | ] ; try (
    simpl in HSq ;
    unfold NetStates in HSq ;
    specialize HSq with i ;
    intuition ).  
  (* i not in I *)
  rewrite <- H3.
  rewrite (Procs_cond TrM i H0).
  rewrite (ST_cond _ (TrG TrM i)).
  left.
  split.
  exact H.
  unfold In, not ; intros.
  apply HNMsi.
  apply Union_introl.
  split ; assumption.
Qed.

Lemma C_related_TauPathL :
  forall s p ph, let I := MatchRIds TrM s in
    forall (BI : BB (SystemLTS (LeftNetwork (RSk) I)) (SystemLTS (RightNetwork (RSk) I))),
      C p s-> (forall j, j < n -> ~I j -> nthS j p = nthS j ph) ->
        (exists sm pm phm, MapR TrM I sm s /\ MapL TrM I pm p /\ MapL TrM I phm ph 
            /\ TauPath (SystemLTS (LeftNetwork RS I)) pm phm /\ BI phm sm) -> C ph s.
Proof.
intros.
unfold I in *.
destruct H1 as [sm [pm [phm [HMsm [HMpm [HMphm [HTR HB]]]]]]].
unfold C, C_rel, D_rel, Bhat_rel in *.
intros i Hin.
rewrite <- (TauPath_MatchL_prop TrM I p ph pm phm HMpm HMphm H0 HTR).
rewrite (C_related_NetMatchIds_prop _ _ H) in H |- *.
rewrite (Union_same (MatchRIds TrM s)) in *.
split ; [intro HNMi | intro HMi].
  (* i is not in M(p) *)
  split.
  destruct HMphm as [HSph _].
  apply HSph ; exact Hin.
  rewrite <- (H0 i Hin HNMi).
  destruct (H i Hin) as [HC _].
  destruct (HC HNMi) as [_ HRW].
  exact HRW.
  (* i is in M(s) *)
  exists phm, sm, BI.
  intuition.
Qed.

Lemma C_related_TauPathR :
  forall s p ph, let I := MatchLIds TrM s in
    forall (BI : BB (SystemLTS (LeftNetwork (RSk) I)) (SystemLTS (RightNetwork (RSk) I))),
      C s p -> (forall j, j < n -> ~I j -> nthS j p = nthS j ph) ->
        (exists sm pm phm, MapL TrM I sm s /\ MapR TrM I pm p /\ MapR TrM I phm ph 
            /\ TauPath (SystemLTS (RightNetwork RS I)) pm phm /\ BI sm phm) -> C s ph.
Proof.
intros.
unfold I in *.
destruct H1 as [sm [pm [phm [HMsm [HMpm [HMphm [HTR HB]]]]]]].
unfold C, C_rel, D_rel, Bhat_rel in *.
intros i Hin.
rewrite <- (TauPath_MatchR_prop TrM I p ph pm phm HMpm HMphm H0 HTR).
rewrite <- (C_related_NetMatchIds_prop _ _ H) in H |- *.
rewrite (Union_same (MatchLIds TrM s)) in *.
split ; [intro HNMi | intro HMi].
  (* i is not in M(s) *)
  split.
  destruct HMsm as [HSs _].
  apply HSs ; exact Hin.
  rewrite <- (H0 i Hin HNMi).
  destruct (H i Hin) as [HC _].
  destruct (HC HNMi) as [_ HRW].
  exact HRW.
  (* i is in M(s) *)
  exists sm, phm, BI.
  intuition.
Qed.

Lemma C_relate_TransL :
  forall v s s' sm s'm p pm, let I := MatchLIds TrM s in
    forall B : BB (SystemLTS (LeftNetwork RSk I)) (SystemLTS (RightNetwork RSk I)),
      MapL TrM I sm s -> MapL TrM I s'm s' -> MapR TrM I pm p -> 
        NetTransSL M v s s' -> NetTransSL (LeftNetwork RS I) v sm s'm -> 
          MatchLIds TrM s = MatchRIds TrM p -> (forall i, i < n -> ~I i -> nthS i s = nthS i p) -> 
            B s'm pm -> C s' p.
Proof.
  intros v s s' sm s'm p pm I B HMsm HMs'm HMpm HT HTL HCIds Heq_sp HB i Hin.
  assert (Trans (SystemLTS M) (act v) s s') as HT2 by (simpl ; exists v ; intuition).
  destruct (trans_cond _ _ _ _ HT2) as [HSs [HSs' _]].
  pose proof (Trans_MatchL_prop TrM _ _ _ _ _ _ HMsm HMs'm HT HTL) as HMss'.
  unfold C_rel, Bhat_rel ; rewrite <- HMss', <- HCIds, (Union_same (MatchLIds TrM s)).
  split ; [intro HMsi | intro HMsi].
    (* i is not in M(s) *)
    split ; try (apply HSs' ; exact Hin).
    pose proof (Ac_Matched_Trans_MatchLIds_prop TrM _ _ _ _ _ HMsm 
                 (NetTranSL_or_intror _ _ _ _ HTL) i Hin HMsi) as HAci.
    rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT i Hin HAci).
    apply Heq_sp ; assumption.
    (* j is in M(s') *)
    exists s'm, pm, B.
    intuition.
Qed.

Lemma C_relate_TransR :
  forall v s s' sm s'm p pm, let I := MatchRIds TrM s in
    forall B : BB (SystemLTS (LeftNetwork RSk I)) (SystemLTS (RightNetwork RSk I)),
      MapR TrM I sm s -> MapR TrM I s'm s' -> MapL TrM I pm p -> 
        NetTransSL TrM v s s' -> NetTransSL (RightNetwork RS I) v sm s'm -> 
          MatchLIds TrM p = MatchRIds TrM s -> (forall i, i < n -> ~I i -> nthS i p = nthS i s) -> 
            B pm s'm -> C p s'.
Proof.
  intros v s s' sm s'm p pm I B HMsm HMs'm HMpm HT HTL HCIds Heq_sp HB i Hin.
  assert (Trans (SystemLTS TrM) (act v) s s') as HT2 by (simpl ; exists v ; intuition).
  destruct (trans_cond _ _ _ _ HT2) as [HSs [HSs' _]].
  pose proof (Trans_MatchR_prop TrM _ _ _ _ _ _ HMsm HMs'm HT HTL) as HMss'.
  unfold C_rel, Bhat_rel ; rewrite <- HMss', <- HCIds, (Union_same (MatchLIds TrM p)).
  split ; [intro HMsi | intro HMsi].
    (* i is not in M(s) *)
    pose proof HMsi as HMpi.
    rewrite HCIds in HMsi.
    pose proof HMpm as [HSp _].
    split ; try (apply HSp ; exact Hin).
    pose proof (Ac_Matched_Trans_MatchRIds_prop TrM _ _ _ _ _ HMsm 
                 (NetTranSL_or_intror _ _ _ _ HTL) i Hin HMsi) as HAci.
    rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT i Hin HAci).
    rewrite Heq_sp ; try assumption ; reflexivity.
    (* j is in M(p') *)
    rewrite HCIds.
    exists pm, s'm, B.
    intuition.
Qed.

Lemma C_rel_branching :
  cond1_universal_applicabilityL TrM -> 
  cond3_completenessL RS ->
  RuleSystem_cascading RSk -> PatternNetworks_bisimilar RSk ->
    forall s p, C s p ->
      forall a s', Trans (SystemLTS M) a s s' -> 
             (a = tau /\ C s' p) 
          \/ (exists ph p', TauPath (SystemLTS TrM) p ph /\ Trans (SystemLTS TrM) a ph p' /\ C s ph /\ C s' p').
Proof.
intros HSyncCond1 HSyncCond3 HRuleSystem_cascading HPatternNetworks_bisimilar 
       s p HC a s' HT.
pose proof (C_related_NetMatchIds_prop s p HC) as HCIds.
pose proof (C_related_NetMatchIds_i_prop s p HC) as HCIdsi.
pose proof HC as HC_rel.
unfold C, C_rel, D_rel, Bhat_rel in HC.
destruct (trans_cond _ a s s' HT) as [HSs [HSs' _]].
unfold C.
destruct HT as [v [Hva HT]].
destruct (C_rel_states _ _ HC) as [_ HSp].
rewrite <- HCIds, (Union_same (MatchLIds TrM s)) in HC.
assert (forall i, i < n -> ~MatchLIds TrM s i -> nthS i s = nthS i p) as HCsp by
  ( intros i Hin HMsi ; destruct (HC i Hin) as [HC' _] ; intuition ).

destruct (classic (forall i, i < n -> Ac v i -> ~MatchLIds TrM s i)) as [HC1 | HC1].
  (* Case: Ac part of s is not removed *)
  assert (forall i, i < n -> Ac v i -> nthS i s = nthS i p) as HAc_sp by
   ( intros ; apply HCsp ; try apply HC1 ; assumption ).
  right.
  pose proof (Cmps_In p s' (Ac v)) as Hp'.
  remember (Cmps p s' (Ac v)) as p'.
  exists p, p'.
  split.
  apply rt_refl (* TauPath TrM p p *).
  assert (NetTransSL TrM v p p') as HTTrM.
    destruct HT as [HVv HT].
    split.
    rewrite SyncN_cond.
    left.
    exact HVv.
    intros i Hin.
    rewrite (Procs_cond TrM i Hin).
    specialize HT with i.
    specialize Hp' with i.
    destruct Hp' as [Hp'Some Hp'None] ; try exact Hin.
    case_eq (nth i (sync v) None) ; [intros a0 Hv | intros Hv] ; rewrite Hv in HT.
      (* i in Ac(v) *)
      apply (TT_cond (mLs TrM i) (TrG TrM i) a0 (nthS i p) (nthS i p')).
      pose proof (AcSome_prop _ _ _ Hv) as HAcvi.
      unfold not.
      left.
      rewrite <- (HAc_sp i Hin HAcvi).
      unfold nthS.
      rewrite (Hp'Some HAcvi).
      split ; try exact (HT Hin).
      intro HM.
      apply (HC1 i) ; try assumption.
      split ; try exact Hin.
      destruct HM as [q [_ [Hmq _]]].
      exists q ; exact Hmq.
      (* i not in Ac(v) *)
      apply AcNone_prop in Hv.
      unfold nthS.
      rewrite Hp'None ; try exact Hv.
      split ; try reflexivity.
      simpl in HSp ; unfold NetStates in HSp ; specialize HSp with i.
      rewrite (Procs_cond TrM i Hin) in HSp.
      split ; exact (HSp Hin)
  (* End assert Trans *).
  assert (forall i, Ac v i -> forall q', (mLs TrM i) q' (nthS i s') -> 
                                Init (L (nth i (Rules RS) d_tr)) q') as HILAc'.
    intros i Hv q' Hmq'.
    pose proof (Acv_i_st_n v i Hv) as Hin.
    pose proof Hv as [a' Hva'].
    apply (TransPartlyMatchedL_prop TrM v a' _ _ q' i HT Hva') ; right.
    split ; try exact Hmq'.
    intro q ; left.
    apply MatchedLIds_not_prop ; try exact Hin.
    unfold not in HC1 |- * ; intro HMpi.
    apply HC1 with (i := i) ; try assumption
  (* End assert Init *).
  pose proof (ComposedToState_MatchLIds TrM v s s' p p' HT Heqp' HCIdsi HILAc') as HCIds'.
  split.
  exists v.
  split ; try exact Hva.
  exact HTTrM.
  unfold C_rel, D_rel, Bhat_rel.
  rewrite <- HCIds, (Union_same (MatchLIds TrM s)).
  rewrite <- HCIds', (Union_same (MatchLIds TrM s')).
  split ; try exact HC.
  (* s' C p' *)
  intros i Hin.
  split ; intro HI'.
    (* i not in I *)
    specialize HC with i.
    split ; auto.
    destruct (Hp' i) as [Hp'Some Hp'None] ; [exact Hin | ].
    destruct (EnsembleNat_dec (Ac v) i) as [Hv | Hv].
      (* i in Ac(v) *)     unfold nthS ; rewrite Hp'Some ; intuition.
      (* i not in Ac(v) *) rewrite <- (NetTransSL_i_not_in_Ac M v _ _ HT i Hin Hv).
      unfold nthS ; rewrite (Hp'None Hv).
      destruct HC as [HC _] ; [exact Hin | ].
      apply HC.
      rewrite (NotAcMatchL_prop TrM v _ _ HT i Hin Hv).
      exact HI'.
    (* i in I *)
    destruct (exists_mapLState TrM (Intersection _ (MatchLIds TrM s') (Ac v)) s' HSs') as [s'm [HMs'm HNIs'm]].
      intros j HIj ; destruct HIj as [j HIj Hv] ; assumption.
    assert (forall j, MatchLIds TrM s' j -> Ac v j -> Init (L (nth j (Rules RS) d_tr)) (nthS j s'm)) as HIs'm2.
      intros j HMs'j HAcj.
      destruct HMs'm as [_ [_ HM]].
      pose proof (Acv_i_st_n v j HAcj) as Hjn.
      apply (HILAc' j HAcj (nthS j s'm)).
      apply (HM j Hjn).
      split ; [exact HMs'j | exact HAcj]
      (* End assert Init *).
    assert (Init (SystemLTS (LeftNetwork RS (Intersection _ (MatchLIds TrM s') (Ac v)))) s'm) as HIs'm.
      apply LeftNetwork_Init_lift.
      intros j Hjn ; split ; intro HI'j.
        (* j in I' *) destruct HI'j as [j HI'j Hv] ; apply HIs'm2 ; assumption.
        (* j not in I' *) exact (DummyLTS_Init s'm j (HNIs'm j Hjn HI'j))
      (* End assert Init *).
    destruct (classic (forall j, ~Ac v j -> ~MatchLIds TrM s' j)) as [HC2 | HC2].
      (* Intersect I' Ac(v) = I *)
      pose proof (contrapositive_from_not_1 (Ac v) (MatchLIds TrM s') HC2) as HI'Ac.
      assert (Included _ (MatchLIds TrM s') (Ac v)) as HI'inAcv by intuition.
      rewrite (Included_Intersection (MatchLIds TrM s') (Ac v) HI'inAcv) in HIs'm, HMs'm.
      destruct (HPatternNetworks_bisimilar (MatchLIds TrM s')) as [BI' _].
      exists s'm, s'm, BI'.
      split ; try assumption.
      split.
        (* s'm maps on p' *)
        split.
          (* p' is a state in the transformed network *)
          simpl ; unfold NetStates ; intros j Hjn.
          rewrite (Procs_cond TrM j Hjn).
          destruct (Hp' j Hjn) as [Hp'S Hp'N].
          destruct (EnsembleNat_dec (Ac v) j) as [Hv | Hv].
            (* Ac v j *)
            unfold nthS.
            rewrite (Hp'S Hv).
            destruct (EnsembleNat_dec (MatchLIds TrM s') j) as [HI'j | HI'j].
              (* s'[j] is matched *)
              pose proof (HIs'm2 j HI'j Hv) as HIs'mj.
              destruct HMs'm as [_ [_ Hms'm]].
              pose proof (Hms'm j Hjn HI'j) as Hms'mj.
              rewrite (mR_init_cond _ (TrG TrM j) _ HIs'mj) in Hms'mj.
              destruct (m_states _ _ _ _ _ Hms'mj) as [_ HSs'mj].
              exact HSs'mj.
              (* s'[j] not is matched *)
              rewrite (ST_cond _ (TrG TrM j)).
              left.
              unfold SetminusS, Setminus, image_mr, In, not.
              split.
              apply HSs' ; exact Hjn.
              intros ; destruct H as [q Hmq].
              apply HI'j ; split ; try exact Hjn.
              exists q ; exact Hmq.
            (* ~Ac V j *)
            unfold nthS.
            rewrite (Hp'N Hv).
            simpl in HSp ; unfold NetStates in HSp ; specialize HSp with j.
            rewrite (Procs_cond TrM j Hjn) in HSp.
            exact (HSp Hjn).
          split.
            (* s'm is a state in the right network *)
            apply init_cond.
            rewrite <- (Net_Init_L_to_Init_R RS (MatchLIds TrM s')).
            exact HIs'm.
          (* forall i in I', s'm[i] matches on p'[i] *)
          intros j Hjn HIj.
          rewrite <- (mR_init_cond _ (TrG TrM j) _ (HIs'm2 j HIj (HI'Ac j HIj))).
          destruct (Hp' j) as [Hp'S _] ; try exact Hjn.
          unfold nthS.
          rewrite Hp'S ; try exact (HI'Ac j HIj).
          destruct HMs'm as [_ [_ Hms']].
          apply Hms' ; assumption.
        (* s'm BI s'm *)
        exact (BI_rel_glue_states_related _ _ BI' s'm HIs'm).
      (* There is a j in 1..n \ Ac(v) such that j in I' *)
      apply not_all_not_exD1 in HC2.
      destruct HC2 as [j [HAcj HI'j]].
      assert (j < n) as Hjn by (destruct HI'j as [Hjn _] ; exact Hjn).
      assert (MatchLIds TrM s j) as HIj by ( unfold MatchLIds, MatchIds in HI'j |- * ; 
        rewrite (NetTransSL_i_not_in_Ac M v s s' HT j Hjn HAcj) ; exact HI'j ).
      destruct (HC j Hjn) as [_ HM] ; clear HC.
      destruct (HM HIj) as [sm [pm [BI [HMsm [HMpm HBI]]]]] ; clear HM HI'j HAcj Hjn HIj j.
      assert (Included nat (MatchLIds TrM s) (MatchLIds TrM s')) as HMss'.
        unfold Included, In ; intros j HMsj.
        pose proof HMsj as [Hjn _].
        destruct (EnsembleNat_dec (Ac v) j) as [HAcj | HAcj].
          (* x in Ac(v) *) pose proof (HC1 j Hjn HAcj) ; contradiction.
          (* x not in Ac(v) *) rewrite <- (NotAcMatchL_prop TrM _ _ _ HT j Hjn HAcj) ; exact HMsj
      (* End assert *).
      assert (Union state_type (MatchLIds TrM s) (Intersection state_type (MatchLIds TrM s') (Ac v)) 
                = (MatchLIds TrM s')) as HIJ.
        apply Extensionality_Ensembles ; unfold Same_set, Included, In.
        split ; intros x Hx.
          destruct Hx as [x Hx | x Hx].
            (* x is in M(s) *)           apply HMss' ; exact Hx.
            (* x is in M(s') and Ac v *) destruct Hx as [x Hx _] ; exact Hx.
          destruct (EnsembleNat_dec (Ac v) x) as [Hv | Hv].
            (* Ac v x *)  right ; split ; [ exact Hx | exact Hv ].
            (* ~Ac v x *) left ; unfold In, MatchLIds, MatchIds in Hx |- * ; pose proof Hx as [Hxn _].
            rewrite (NetTransSL_i_not_in_Ac M v s s' HT x Hxn Hv) ; exact Hx
      (* End assert Union ... *).
      destruct (HPatternNetworks_bisimilar (Intersection _ (MatchLIds TrM s') (Ac v))) as [BJ _].
      pose proof (BI_rel_glue_states_related _ _ BJ _ HIs'm) as HBJ.
      assert (Disjoint nat (MatchLIds TrM s) (Intersection _ (MatchLIds TrM s') (Ac v))) as HDis by
        ( apply Disjoint_intro ; unfold In, not ; intros x Hx ;
          destruct Hx as [x HIx HI'Acx] ;
          destruct HI'Acx as [x HI'x HAcx] ;
          pose proof (Acv_i_st_n v x HAcx) as Hxn ;
          apply (HC1 x) ; assumption ).
      pose proof (HRuleSystem_cascading _ _ HDis sm pm s'm BI) as HCas.
      rewrite HIJ in HCas.
      destruct (HPatternNetworks_bisimilar (MatchLIds TrM s')) as [BIJ _].
      destruct (HCas BIJ HIs'm) as [HBIJ _] ; clear HCas.
      exists (Cmps sm s'm (Intersection state_type (MatchLIds TrM s') (Ac v))),
             (Cmps pm s'm (Intersection state_type (MatchLIds TrM s') (Ac v))), BIJ.
      destruct HMs'm as [_ [_ Hms'm]].
      split.
        (* Composed s'm maps on s' *)
        destruct HMsm as [_ [_ Hmsm]].
        split ; try exact HSs'.
        split ; try ( destruct (BI_rel_states _ _ BIJ _ _ (HBIJ HBI)) as [HS _] ; exact HS ).
        intros j Hjn HMj.
        destruct (Cmps_In sm s'm (Intersection state_type (MatchLIds TrM s') (Ac v)) j Hjn) as [Hs'm Hsm].
        destruct (EnsembleNat_dec (Ac v) j) as [Hv | Hv].
          (* Ac v j *)  unfold nthS ; rewrite Hs'm ; [ apply (Hms'm j Hjn) | ] ; split ; assumption.
          (* ~ Ac v j*) rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn Hv).
          assert (~Intersection state_type (MatchLIds TrM s') (Ac v) j) as HI'j by
            ( apply Intersection_DeMorgan ; right ; exact Hv ).
          unfold nthS ; rewrite (Hsm HI'j).
          apply Hmsm ; try assumption.
          unfold MatchLIds, MatchIds in HMj |- *.
          rewrite (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn Hv) ; exact HMj.
      split.
        (* composed p'm maps on p' *)
        destruct HMpm as [_ [_ Hmpm]].
        split.
          assert (Trans (SystemLTS TrM) a p p') as HTTrM2 by (exists v ; split ; assumption).
          destruct (trans_cond _ _ _ _ HTTrM2) as [_ [HSp' _]] ; exact HSp'.
        split ; try ( destruct (BI_rel_states _ _ BIJ _ _ (HBIJ HBI)) as [_ HS] ; exact HS ).
        intros j Hjn HMj ; destruct (Hp' j Hjn) as [Hp'S Hp'N].
        destruct (Cmps_In pm s'm (Intersection state_type (MatchLIds TrM s') (Ac v)) j Hjn) as [Hs'm Hpm].
        destruct (EnsembleNat_dec (Ac v) j) as [Hv | Hv].
          (* Ac v j *)  unfold nthS ; rewrite Hs'm ; 
          try (pose proof (mR_init_cond _ (TrG TrM j) _ (HIs'm2 j HMj Hv)) as HX ; 
                 unfold nthS in HX ; rewrite <- HX) ;
          [ rewrite (Hp'S Hv) ; apply (Hms'm j Hjn) | ] ; split ; assumption.
          (* ~ Ac v j*) rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HTTrM j Hjn Hv).
          assert (~Intersection state_type (MatchLIds TrM s') (Ac v) j) as HI'j by
            ( apply Intersection_DeMorgan ; right ; exact Hv ).
          unfold nthS ; rewrite (Hpm HI'j).
          apply Hmpm ; try assumption.
          unfold MatchLIds, MatchIds in HMj |- *.
          rewrite (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn Hv) ; exact HMj.
        (* BJ s'm p'm *)
        exact (HBIJ HBI).

  (* Case: A substate of Ac part of s is removed *)
  apply not_all_not_exD2 in HC1.
  destruct (classic (exists i, i < n /\ Ac v i /\
     exists q q', mLs TrM i q (nthS i s) /\ mLs TrM i q' (nthS i s') /\ 
        exists ai, nth i (sync v) None = Some ai /\ Trans (L (nth i (Rules RS) d_tr)) ai q q')) as [HC2 | HC2].

    (* Case: There is a matching transition *)
    destruct HC2 as [i [Hin [HAci HMTrans]]].
    assert (MatchLIds TrM s i) as HMsi.
      destruct HMTrans as [q [_ [Hmq _]]].
      split ; try exact Hin.
      exists q ; exact Hmq
    (* End assert M(s) i *).
    destruct (HC i Hin) as [_ HCM].
    destruct (HCM HMsi) as [sm [pm [B [HMsm [HMpm HB]]]]] ; clear HCM.
    destruct (Ac_MappedL_prop TrM HSyncCond1 HSyncCond3 v s s' sm i HT HMsm HMTrans) as [s'm [HMs'm HTL]].
    assert (Trans (SystemLTS (LeftNetwork RS (MatchLIds TrM s))) a sm s'm)
      as HTL2 by ( simpl ; exists v ; split ; try exact Hva ; exact HTL ).
    destruct (BI_rel_branching _ _ B _ _ HB _ _ HTL2) as [HTB | HTB].
      (* Case: a = tau /\ B s'm pm *)
      left.
      destruct HTB as [Ha HB2].
      split ; try exact Ha.
      (* C s' p *) exact (C_relate_TransL v s s' sm s'm p pm B HMsm HMs'm HMpm HT HTL HCIds HCsp HB2).
      (* Case: pm ==> phm -a-> p'm *)
      right.
      destruct HTB as [phm [p'm [HTPR [HTR [HBh HB']]]]].
      destruct (MapR_TauPath TrM (MatchLIds TrM s) _ _ HTPR _ HMpm) as [ph [HMphm [HTP2 Heqph]]].
      destruct HTR as [v' [Hv' HTR]].
      destruct (MapR_trans TrM (MatchLIds TrM s) _ _ _ _ HTR HMphm) as [p' [HMp'm [HT2 Heqp']]].
      exists ph, p'.
      split ; try exact HTP2.
      split ; try (exists v' ; split ; assumption).
      assert (C s ph) as HCsph.
        unfold C, C_rel, Bhat_rel.
        rewrite <- (TauPath_MatchR_prop _ _ _ _ _ _ HMpm HMphm Heqph HTPR).
        rewrite <- HCIds, (Union_same (MatchLIds TrM s)).
        intros j Hjn.
        split ; intro HMsj.
        split ; [apply HSs ; exact Hjn | ].
        rewrite <- (Heqph j Hjn HMsj).
        apply HCsp ; assumption.
        exists sm, phm, B ; intuition
        (* End assert C s ph *).
      split ; try exact HCsph.
      (* C s' p' *) 
      rewrite (TauPath_MatchR_prop _ _ _ _ _ _ HMpm HMphm Heqph HTPR),
              (Trans_MatchR_prop _ _ _ _ _ _ _ HMphm HMp'm HT2 HTR) in HCIds.
      apply (C_relate_TransL v s s' sm s'm p' p'm B HMsm HMs'm HMp'm HT HTL HCIds) ; intuition.
      rewrite <- Heqp', <- Heqph ; try apply HCsp ; try assumption.

    (* Case: There is no matching transition *)
    right.
    assert (forall i, i < n -> Ac v i -> 
      ~(exists q q', (mLs TrM i) q (nthS i s) /\ (mLs TrM i) q' (nthS i s') 
           /\ exists ai, nth i (sync v) None = Some ai /\ Trans (L (nth i (Rules RS) d_tr)) ai q q')) 
      as HC3 by ( apply not_ex_all_notD ; apply HC2 ) ; clear HC2.
    destruct HC1 as [i [Hin [HAci HMsi]]].
    pose proof HAci as [ai Hai].
    destruct (HC i Hin) as [_ HC'].
    destruct (HC' HMsi) as [sm [pm [B [HMsm [HMpm HB]]]]] ; clear HC'.
    pose proof HMsm as [_ [HSsm _]].
    assert (forall j, j < n -> MatchLIds TrM s j ->
              Ac v j -> Init (L (nth j (Rules RS) d_tr)) (nthS j sm)) as HIAcsm.
      intros j Hjn HMsj HAcj.
      pose proof HAcj as [aj Haj].
      destruct HMsm as [_ [_ Hmsm]].
      specialize Hmsm with j.
      apply (TransPartlyMatchedL_prop2 TrM _ s s' (nthS j sm) j HT HAcj) ;
      try (left ; apply Hmsm) ; assumption
    (* End assert Init *).
    destruct (TauPath_glue_state_L_prop RSk (MatchLIds TrM s) (Ac v) B sm pm HB HIAcsm) 
      as [phm [HTPR [HBphm Heqphm]]].
    destruct (MapR_TauPath TrM (MatchLIds TrM s) pm phm HTPR p HMpm) as [ph [HMphm [HTP Heqph]]].
    assert (forall j, j < n -> MatchLIds TrM s j -> Ac v j -> nthS j s = nthS j ph) as Heqph2.
      intros j Hjn HIj HAcj.
      apply (m_injective _ _ (mLs TrM j) _ _ (nthS j sm)).
      destruct HMsm as [_ [_ Hmsm]].
      destruct HMphm as [_ [_ Hmphm]].
      left ; split.
      apply Hmsm ; assumption.
      rewrite (mR_init_cond _ (TrG TrM j) _ (HIAcsm j Hjn HIj HAcj)).
      unfold nthS.
      rewrite <- (Heqphm j Hjn HIj HAcj).
      apply Hmphm ; assumption
    (* End assert *).
    pose proof (Cmps_In ph s' (Ac v)) as Hp'.
    remember (Cmps ph s' (Ac v)) as p'.
    assert (NetTransSL TrM v ph p') as HT'.
      destruct HT as [Hv HT].
      split.
      rewrite SyncN_cond.
      left ; exact Hv.
      intros j Hjn.
      specialize HT with j.
      destruct (Hp' j Hjn) as [Hp'S Hp'N].
      case_eq (nth j (sync v) None) ; intros ; rewrite H in HT ; unfold nthS.
        (* v[j] = a0 *)
        unfold nthS in Heqph2, Heqph.
        pose proof (AcSome_prop _ _ _ H) as HAcj.
        rewrite (Hp'S HAcj).
        destruct (EnsembleNat_dec (MatchLIds TrM s) j) as [HMsj | HNMsj] ;
          [ (* j in M(s) *)
            rewrite <- (Heqph2 j Hjn HMsj HAcj)
          | (* j not in M(s) *)
            rewrite <- (Heqph j Hjn HNMsj) ;
            destruct (HC j Hjn) as [HC' _] ;
            unfold nthS in HC' ;
            destruct (HC' HNMsj) as [_ HRW] ;
            rewrite <- HRW ] ;
          rewrite (Procs_cond TrM) ; try exact Hjn ;
          apply (TT_cond _ (TrG TrM j) a0 (nthS j s) (nthS j s')) ;
          left ; split ; try (apply HT ; assumption) ;
          unfold not ; intro HTL ;
          destruct HTL as [q [q' [Hmq [Hmq' HTL]]]] ;
          apply (HC3 j Hjn HAcj) ;
          exists q, q' ;
          split ; try split ; try exists a0 ;
          try split ; assumption.
        (* v[j] = * *)
        apply AcNone_prop in H.
        rewrite (Hp'N H).
        destruct HMphm as [HSph _].
        simpl in HSph.
        unfold NetStates in HSph.
        specialize HSph with j.
        intuition
    (* End assert Trans *).
    assert (Trans (SystemLTS TrM) a ph p') as HT'2 by (exists v ; split ; assumption).
    exists ph, p'.
    split ; try exact HTP.
    split ; try (simpl ; exists v ; split ; assumption).
    assert (C s ph) as HCsph 
      by (apply (C_related_TauPathR s p ph B HC_rel Heqph) ; exists sm, pm, phm ; intuition).
    split.
      (* C s ph *) exact HCsph.
      (* C s' p' *)
      unfold C_rel, D_rel, Bhat_rel ; clear Hin HAci HMsi Hai ai i.
      assert (forall i, Ac v i -> forall q', (mLs TrM i) q' (nthS i s') -> 
                Init (L (nth i (Rules RS) d_tr)) q') as HILAc'.
        intros i Hv q' Hmq'.
        pose proof (Acv_i_st_n v i Hv) as Hin.
        pose proof Hv as [a' Hva'].
        apply (TransPartlyMatchedL_prop TrM v a' _ _ q' i HT Hva') ; right.
        split ; try exact Hmq'.
        intro q ; apply not_and_or.
        unfold not in HC3 |- * ; intro HTm.
        apply HC3 with (i := i) ; try assumption.
        exists q, q' ; intuition.
        exists a' ; intuition
        (* End assert Init *).
      pose proof (ComposedToState_MatchLIds TrM v s s' ph p' HT Heqp' 
                    (C_related_NetMatchIds_i_prop s ph HCsph) HILAc') as HCIds'.
      rewrite <- HCIds', (Union_same (MatchLIds TrM s')).
      intros i Hin.
      split ; [intro HMs'i | intro HMs'i].
        (* i is not in M(s) *)
        split.
        apply HSs' ; exact Hin.
        destruct (Hp' i Hin) as [Hp'S Hp'N].
        destruct (EnsembleNat_dec (Ac v) i) as [HAci | HAci].
          (* Ac v i *)
          unfold nthS.
          rewrite (Hp'S HAci).
          reflexivity.
          (* ~Ac v i *)
          rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT i Hin HAci).
          rewrite <- (NotAcMatchL_prop TrM _ _ _ HT i Hin HAci) in HMs'i.
          unfold nthS.
          rewrite (Hp'N HAci).
          destruct (HC i Hin) as [HC' _].
          destruct (HC' HMs'i) as [_ HRW].
          rewrite (Heqph i Hin HMs'i) in HRW.
          apply HRW.
        (* i is in M(s') *)
        set (I := Intersection nat (MatchLIds TrM s) (Ac v)).
        set (J := Setminus nat (MatchLIds TrM s') (Ac v)).
        assert (Disjoint nat J I) as HDis_J_I by (unfold I, J ; apply Disjoint_Intersection_Setminus).
        assert (Setminus nat (MatchLIds TrM s) (Ac v) = Setminus nat (MatchLIds TrM s') (Ac v)) as HJ.
          apply Extensionality_Ensembles.
          unfold Same_set, Included, In ; split ; intros x Hx ;
          destruct Hx as [HMsmx HAcx] ; pose proof HMsmx as [Hxn _] ;
          split ; unfold In ;
          solve [ rewrite <- (NotAcMatchL_prop TrM _ _ _ HT x Hxn HAcx) ; exact HMsmx
                | rewrite (NotAcMatchL_prop TrM _ _ _ HT x Hxn HAcx) ; exact HMsmx
                | exact HAcx ]
          (* End assert Setminus... *).
        pose proof (Cmps_In (d_staten n) sm J) as HsmJ.
        remember (Cmps (d_staten n) sm J) as smJ.
        pose proof (Cmps_In (d_staten n) phm J) as HphmJ.
        remember (Cmps (d_staten n) phm J) as phmJ.
        pose proof (Cmps_In (d_staten n) sm I) as HsmI.
        remember (Cmps (d_staten n) sm I) as smI.
        assert (NetInit (LeftNetwork RS I) smI) as HIsmI.
          apply LeftNetwork_Init_lift.
          unfold nthS.
          intros x Hxn ; split ; intro HIx ; 
            [destruct (HsmI x Hxn) as [HsmI2 _] | destruct (HsmI x Hxn) as [_ HsmI2] ] ; rewrite (HsmI2 HIx).
          destruct HIx ; apply HIAcsm ; assumption.
          unfold d_state ; rewrite d_staten_nth ; reflexivity
          (* End assert NetInit *).
        destruct (HPatternNetworks_bisimilar J) as [BJ _].
        pose proof (HRuleSystem_cascading J I HDis_J_I smJ phmJ smI BJ) as HCas.
        assert ((MatchLIds TrM s) = (Union nat J I)) as HU.
          apply Extensionality_Ensembles.
          unfold I, J, Same_set, Included, In.
          rewrite <- HJ.
          split ; intros x Hx.
          (* x in M(s) *)
          destruct (classic ((Ac v) x)) as [HAcx | HAcx ].
            (* x in Ac *) apply Union_intror ; apply Intersection_intro ; assumption.
            (* x not in Ac *) apply Union_introl ; split ; assumption.
          (* x not in M(s) *)
          destruct Hx as [x Hx | x Hx].
            (* x in J *) destruct Hx as [HMsx HAcx] ; exact HMsx.
            (* x in I *) destruct Hx ; assumption
          (* End assert MLids s = Union J I *).
        rewrite <- HU in HCas ; clear HU.
        destruct (HCas B HIsmI) as [_ HBJ] ; clear HCas.
        rewrite HeqsmJ, HeqsmI, HeqphmJ in HBJ.
        unfold I, J in HBJ.
        assert (forall x, I x -> nth x sm dummy = nth x sm dummy) as Hreq1 by (intros ; reflexivity).
        assert (forall x, x < n -> ~ MatchLIds TrM s x -> nthS x sm = d_state) as Hreq2 
          by (intros x Hxn HMsx ; apply (LeftNetwork_States_simpl _ _ _ x HSsm Hxn) ; exact HMsx).
        pose proof (Cas_Cmps_state_prop v s sm sm _ Hreq1 Hreq2) as Hrw ; rewrite HJ in Hrw.
        rewrite Hrw in HBJ ; clear Hreq1 Hreq2 Hrw.
        assert (forall x, I x -> nth x sm dummy = nth x phm dummy) as Hreq1 
          by ( intros x HIx ; destruct HIx as [x HMsmx HAcx] ; pose proof HMsmx as [Hxn _] ;
               rewrite (Heqphm x Hxn HMsmx HAcx) ; reflexivity ).
        assert (forall x, x < n -> ~ MatchLIds TrM s x -> nthS x phm = d_state) as Hreq2 
          by ( intros x Hxn HMsx ; destruct (B_states _ _ _ _ _ HBphm) as [_ HSphm] ;
               destruct (RightNetwork_States_simpl _ _ _ x HSphm Hxn) as [_ Hphmx] ; apply Hphmx ; assumption ).
        pose proof (Cas_Cmps_state_prop v s sm phm _ Hreq1 Hreq2) as Hrw ; rewrite HJ in Hrw.
        rewrite Hrw in HBJ ; clear Hreq1 Hreq2 Hrw.
        unfold I, J in * ; rewrite <- HeqsmJ in HBJ ; rewrite <- HeqphmJ in HBJ.
        clear HIsmI HsmI HeqsmI smI HDis_J_I I.
        set (I := Intersection nat (MatchLIds TrM s') (Ac v)).        
        destruct (exists_mapLState TrM I s' HSs') as [s'mI [HMs'mI HIN]].
          intros j HIj ; destruct HIj as [j HMs'j _] ; exact HMs'j.
        assert (NetInit (LeftNetwork RS I) s'mI) as HIs'mI.
          apply LeftNetwork_Init_lift.
          intros j Hjn.
          split ; intro HIj.
          (* j in I *)
          destruct HIj as [j HMs'j HAcj].
          destruct HMs'mI as [_ [_ Hms'mI]].
          apply HILAc' ; [exact HAcj | apply Hms'mI ; try exact Hjn ; split ; assumption].
          (* j not in I *)
          apply DummyLTS_Init ; apply HIN ; try unfold I ; assumption
        (* End assert *).
        destruct (HPatternNetworks_bisimilar (Union _ J I)) as [BIJ _].
        assert (Disjoint nat J I) as HDis_J_I by (unfold I, J ; apply Disjoint_Intersection_Setminus).
        destruct (HRuleSystem_cascading J I HDis_J_I smJ phmJ s'mI BJ BIJ HIs'mI) as [HCas _].
        pose proof (HCas (HBJ HBphm)) as HBIJ.
        rewrite <- (Union_Setminus_Intersection (MatchLIds TrM s') (Ac v)).
        exists (Cmps smJ s'mI I), (Cmps phmJ s'mI I), BIJ.
        split.
          (* smJ[s'mI / I] maps on s' *)
          split ; try exact HSs'.
          split.
          destruct (BI_rel_states RSk _ _ _ _ HBIJ) as [HSs'm _].
          apply HSs'm.
          intros j Hjn HIJj.
          destruct HIJj as [j HJj | j HIj].
            (* j in J *)
            pose proof HJj as [HMs'j HNAcj] ; unfold In in HMs'j.
            rewrite <- (NetTransSL_i_not_in_Ac M _ _ _ HT j Hjn HNAcj).
            rewrite <- (NotAcMatchL_prop TrM _ _ _ HT j Hjn HNAcj) in HMs'j.
            unfold nthS.
            pose proof (Disjoint_In _ _ HDis_J_I j HJj) as HNIj.
            rewrite (Cmps_i_not_in_I smJ s'mI I j Hjn HNIj).
            destruct (HsmJ j Hjn) as [HsmJ1 _].
            rewrite (HsmJ1 HJj).
            destruct HMsm as [_ [_ Hmsm]].
            apply Hmsm ; assumption.
            (* j in I *)
            unfold nthS.
            rewrite (Cmps_i_in_I smJ s'mI I j Hjn HIj).
            destruct HMs'mI as [_ [_ Hms'mI]].
            apply Hms'mI ; assumption.
          split.
          (* phmJ[s'mI / I] maps on p' *)
          split.
          destruct (trans_cond _ _ _ _ HT'2) as [_ [HSp' _]].
          exact HSp'.
          split.
          destruct (BI_rel_states RSk _ _ _ _ HBIJ) as [_ HSp'm].
          apply HSp'm.
          intros j Hjn HIJj.
          destruct (Hp' j Hjn) as [Hp'S Hp'N].
          destruct HIJj as [j HJj | j HIj].
            (* j in J *)
            unfold nthS.
            pose proof (Disjoint_In _ _ HDis_J_I j HJj) as HNIj.
            rewrite (Cmps_i_not_in_I _ s'mI I j Hjn HNIj).
            destruct (HphmJ j Hjn) as [HphmJ' _].
            rewrite (HphmJ' HJj).
            destruct HJj as [HMs'j HAcj] ; unfold In in HMs'j.
            rewrite (Hp'N HAcj).
            destruct HMphm as [_ [_ Hmphm]].
            rewrite <- (NotAcMatchL_prop TrM _ _ _ HT j Hjn HAcj) in HMs'j.
            apply Hmphm ; assumption.
            (* j in I *)
            assert (Ac v j) as HAcj by (destruct HIj ; assumption).
            unfold nthS.
            rewrite (Cmps_i_in_I phmJ s'mI I j Hjn HIj).
            rewrite (Hp'S HAcj).
            destruct HMs'mI as [_ [_ Hms'mI]].
            destruct (LeftNetwork_Init_simpl RS I s'mI j HIs'mI Hjn) as [HIs'mI2 _].
            pose proof (mR_init_cond _ (TrG TrM j) _ (HIs'mI2 HIj)) as HRW ; unfold nthS in HRW.
            rewrite <- HRW ; clear HRW.
            apply Hms'mI ; assumption.
         (* s' B p' *)
         assumption.
Qed.

Lemma C_rel_branching_sym :
  cond2_universal_applicabilityR TrM -> 
  cond4_completenessR RS ->
  cond5_new_SyncLaw M RS ->
  RuleSystem_cascading RSk -> 
  PatternNetworks_bisimilar RSk ->
    forall s p, C s p ->
      forall a p', Trans (SystemLTS TrM) a p p' -> 
             (a = tau /\ C s p') 
          \/ (exists sh s', TauPath (SystemLTS M) s sh /\ Trans (SystemLTS M) a sh s' /\ C sh p /\ C s' p').
Proof.
intros HSyncCond2 HSyncCond4 HSyncCond5 
       HRuleSystem_cascading HPatternNetworks_bisimilar
       s p HC a p' HT.
pose proof (C_related_NetMatchIds_prop s p HC) as HCIds.
pose proof (C_related_NetMatchIds_i_prop s p HC) as HCIdsi.
pose proof HC as HC_rel.
unfold C, C_rel, D_rel, Bhat_rel in HC.
destruct (trans_cond _ a p p' HT) as [HSp [HSp' _]].
unfold C.
destruct HT as [v [Hva HT]].
destruct (C_rel_states _ _ HC) as [HSs _].
rewrite HCIds, (Union_same (MatchRIds TrM p)) in HC.
assert (forall i, i < n -> ~MatchRIds TrM p i -> nthS i s = nthS i p) as HCsp by
  ( intros i Hin HMsi ; destruct (HC i Hin) as [HC' _] ; intuition ).

destruct (classic (forall i, i < n -> Ac v i -> ~MatchRIds TrM p i)) as [HC1 | HC1].
  (* Case: Ac part of s is not removed *)
  assert (forall i, i < n -> Ac v i -> nthS i s = nthS i p) as HAc_sp by
   ( intros ; apply HCsp ; try apply HC1 ; assumption ).
  right.
  pose proof (Cmps_In s p' (Ac v)) as Hs'.
  remember (Cmps s p' (Ac v)) as s'.
  exists s, s'.
  split.
  apply rt_refl (* TauPath TrM s s *).
  assert (NetTransSL M v s s') as HTM.
    destruct HT as [HVv HT].
    split.
    assert (NetTransSL TrM v p p') as HT' by (split ; assumption).
    rewrite SyncN_cond in HVv.
    (* SyncN M v *)
    apply (unchanged_transitions_v_in_SynN_M TrM HSyncCond2 HSyncCond5 v _ _  HT').
    destruct (sync_cond v) as [i HAcvi].
    pose proof (listn_nth_bound _ _ _ HAcvi) as Hin.
    apply Option_NNone_Some in HAcvi.
    exists i.
    split ; try exact HAcvi.
    unfold not ; intro HM.
    apply (HC1 i Hin HAcvi).
    split ; try exact Hin.
    destruct HM as [q [_ [Hmq _]]].
    exists q ; exact Hmq
    (* End SyncN M v *).
    intros i Hin.
    specialize HT with i.
    rewrite (Procs_cond TrM i Hin) in HT.
    specialize Hs' with i.
    destruct Hs' as [Hs'Some Hs'None] ; try exact Hin.
    case_eq (nth i (sync v) None) ; [intros a0 Hv | intros Hv] ; rewrite Hv in HT.
      (* i in Ac(v) *)
      apply (TT_cond (mLs TrM i) (TrG TrM i) a0 (nthS i p) (nthS i p')) in HT ; try exact Hin.
      pose proof (AcSome_prop _ _ _ Hv) as HAcvi.
      rewrite (HAc_sp i Hin HAcvi).
      unfold nthS ; rewrite (Hs'Some HAcvi).
      destruct HT as [[HT _] | HT].
        (* Trans was not added *) exact HT.
        (* Trans was added *)
        exfalso ;
        apply (HC1 i Hin HAcvi).
        split ; try exact Hin.
        destruct HT as [q [_ [Hmq _]]].
        exists q ; exact Hmq.
      (* i not in Ac(v) *)
      apply AcNone_prop in Hv.
      unfold nthS.
      rewrite Hs'None ; try exact Hv.
      split ; try reflexivity.
      simpl in HSs ; unfold NetStates in HSs ; specialize HSs with i.
      split ; exact (HSs Hin)
  (* End assert Trans *).
  assert (forall i, Ac v i -> forall q', (mR _ (TrG TrM i)) q' (nthS i p') -> 
                                Init (R (nth i (Rules RS) d_tr)) q') as HIRAc'.
    intros i Hv q' Hmq'.
    pose proof (Acv_i_st_n v i Hv) as Hin.
    pose proof Hv as [a' Hva'].
    apply (TransPartlyMatchedR_prop TrM v a' _ _ q' i HT Hva') ; right.
    split ; try exact Hmq'.
    intro q ; left.
    apply MatchedRIds_not_prop ; try exact Hin.
    unfold not in HC1 |- * ; intro HMpi.
    apply HC1 with (i := i) ; try assumption
  (* End assert Init *).
  pose proof (ComposedToState_MatchRIds TrM v p p' s s' HT Heqs' HCIdsi HIRAc') as HCIds'.
  split.
  exists v.
  split ; try exact Hva.
  exact HTM.
  unfold C_rel, D_rel, Bhat_rel.
  rewrite HCIds , (Union_same (MatchRIds TrM p)).
  rewrite HCIds', (Union_same (MatchRIds TrM p')).
  split ; try exact HC.
  (* s' C p' *)
  intros i Hin.
  split ; intro HI'.
    (* i not in I *)
    specialize HC with i.
    destruct (Hs' i) as [Hs'Some Hs'None] ; [exact Hin | ].
    destruct (EnsembleNat_dec (Ac v) i) as [Hv | Hv].
      (* i in Ac(v) *)     
      unfold nthS ; rewrite (Hs'Some Hv).
      split ; try reflexivity.
      pose proof (HSp' i Hin) as HSs'i.
      rewrite Procs_cond in HSs'i ; try exact Hin.
      rewrite (ST_cond _ (TrG TrM i)) in HSs'i.
      apply Union_to_or in HSs'i.
      destruct HSs'i as [HSs'i | HSs'i].
        (* p' was not added *) destruct HSs'i ; assumption.
        (* p' was added *)     exfalso ; apply HI' ; split ; assumption.
      (* i not in Ac(v) *) 
      rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT i Hin Hv).
      unfold nthS ; rewrite (Hs'None Hv).
      destruct HC as [HC _] ; [exact Hin | ].
      apply HC.
      rewrite (NotAcMatchR_prop TrM v _ _ HT i Hin Hv).
      exact HI'.
    (* i in I *)
    destruct (exists_mapRState TrM (Intersection _ (MatchRIds TrM p') (Ac v)) p' HSp') as [p'm [HMp'm HNIp'm]].
      intros j HIj ; destruct HIj as [j HIj Hv] ; assumption.
    assert (forall j, MatchRIds TrM p' j -> Ac v j -> Init (R (nth j (Rules RS) d_tr)) (nthS j p'm)) as HIp'm2.
      intros j HMp'j HAcj.
      destruct HMp'm as [_ [_ HM]].
      pose proof (Acv_i_st_n v j HAcj) as Hjn.
      apply (HIRAc' j HAcj (nthS j p'm)).
      apply (HM j Hjn).
      split ; [exact HMp'j | exact HAcj]
      (* End assert Init *).
    assert (Init (SystemLTS (RightNetwork RS (Intersection _ (MatchRIds TrM p') (Ac v)))) p'm) as HIp'm.
      apply RightNetwork_Init_lift.
      intros j Hjn ; split ; intro HI'j.
        (* j in I' *) destruct HI'j as [j HI'j Hv] ; apply HIp'm2 ; assumption.
        (* j not in I' *) exact (DummyLTS_Init p'm j (HNIp'm j Hjn HI'j))
      (* End assert Init *).
    rewrite <- Net_Init_L_to_Init_R in HIp'm.
    destruct (classic (forall j, ~Ac v j -> ~MatchRIds TrM p' j)) as [HC2 | HC2].
      (* Intersect I' Ac(v) = I *)
      pose proof (contrapositive_from_not_1 (Ac v) (MatchRIds TrM p') HC2) as HI'Ac.
      assert (Included _ (MatchRIds TrM p') (Ac v)) as HI'inAcv by intuition.
      rewrite (Included_Intersection (MatchRIds TrM p') (Ac v) HI'inAcv) in HIp'm, HMp'm.
      destruct (HPatternNetworks_bisimilar (MatchRIds TrM p')) as [BI' _].
      exists p'm, p'm, BI'.
      split.
        (* p'm maps on s' *)
        split.
          (* s' is a state in the original network *)
          simpl in HSp' |- * ; unfold NetStates in HSp' |- *  ; intros j Hjn.
          specialize HSp' with j ; rewrite (Procs_cond TrM j Hjn) in HSp'.
          destruct (Hs' j Hjn) as [Hs'S Hs'N].
          destruct (EnsembleNat_dec (Ac v) j) as [Hv | Hv].
            (* Ac v j *)
            unfold nthS.
            rewrite (Hs'S Hv).
            destruct (EnsembleNat_dec (MatchRIds TrM p') j) as [HI'j | HI'j].
              (* p'[j] is matched *)
              pose proof (HIp'm2 j HI'j Hv) as HIp'mj.
              destruct HMp'm as [_ [_ Hmp'm]].
              pose proof (Hmp'm j Hjn HI'j) as Hmp'mj.
              rewrite <- Init_L_to_Init_R in HIp'mj.
              rewrite <- (mR_init_cond _ (TrG TrM j) _ HIp'mj) in Hmp'mj.
              destruct (m_states _ _ _ _ _ Hmp'mj) as [_ HSp'mj].
              exact HSp'mj.
              (* p'[j] not is matched *)
              rewrite (ST_cond _ (TrG TrM j)) in HSp'.
              destruct (Union_to_or _ _ _ (HSp' Hjn)) as [[HSp'i _] | HSp'i].
                (* p' was not added *) exact HSp'i.
                (* p' was added *)     exfalso ; apply HI'j ; split ; assumption.
            (* ~Ac V j *)
            unfold nthS.
            rewrite (Hs'N Hv).
            apply HSs ; exact Hjn.
          split.
            (* p'm is a state in the right network *)
            apply init_cond ; exact HIp'm.
          (* forall i in I', p'm[i] matches on s'[i] *)
          intros j Hjn HIj.
          specialize HIp'm2 with j ; rewrite <- Init_L_to_Init_R in HIp'm2.
          rewrite (mR_init_cond _ (TrG TrM j) _ (HIp'm2 HIj (HI'Ac j HIj))).
          destruct (Hs' j) as [Hs'S _] ; try exact Hjn.
          unfold nthS.
          rewrite Hs'S ; try exact (HI'Ac j HIj).
          destruct HMp'm as [_ [_ Hmp']].
          apply Hmp' ; assumption.
        split ; try assumption.
        (* s'm BI s'm *)
        exact (BI_rel_glue_states_related _ _ BI' p'm HIp'm).
      (* There is a j in 1..n \ Ac(v) such that j in I' *)
      apply not_all_not_exD1 in HC2.
      destruct HC2 as [j [HAcj HI'j]].
      assert (j < n) as Hjn by (destruct HI'j as [Hjn _] ; exact Hjn).
      assert (MatchRIds TrM p j) as HIj by ( unfold MatchRIds, MatchIds in HI'j |- * ; 
        rewrite (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn HAcj) ; exact HI'j ).
      destruct (HC j Hjn) as [_ HM] ; clear HC.
      destruct (HM HIj) as [sm [pm [BI [HMsm [HMpm HBI]]]]] ; clear HM HI'j HAcj Hjn HIj j.
      assert (Included nat (MatchRIds TrM p) (MatchRIds TrM p')) as HMpp'.
        unfold Included, In ; intros j HMpj.
        pose proof HMpj as [Hjn _].
        destruct (EnsembleNat_dec (Ac v) j) as [HAcj | HAcj].
          (* x in Ac(v) *) pose proof (HC1 j Hjn HAcj) ; contradiction.
          (* x not in Ac(v) *) rewrite <- (NotAcMatchR_prop TrM _ _ _ HT j Hjn HAcj) ; exact HMpj
      (* End assert *).
      assert (Union state_type (MatchRIds TrM p) (Intersection state_type (MatchRIds TrM p') (Ac v)) 
                = (MatchRIds TrM p')) as HIJ.
        apply Extensionality_Ensembles ; unfold Same_set, Included, In.
        split ; intros x Hx.
          destruct Hx as [x Hx | x Hx].
            (* x is in M(s) *)           apply HMpp' ; exact Hx.
            (* x is in M(s') and Ac v *) destruct Hx as [x Hx _] ; exact Hx.
          destruct (EnsembleNat_dec (Ac v) x) as [Hv | Hv].
            (* Ac v x *)  right ; split ; [ exact Hx | exact Hv ].
            (* ~Ac v x *) left ; unfold In, MatchRIds, MatchIds in Hx |- * ; pose proof Hx as [Hxn _].
            rewrite (NetTransSL_i_not_in_Ac TrM v p p' HT x Hxn Hv) ; exact Hx
      (* End assert Union ... *).
      destruct (HPatternNetworks_bisimilar (Intersection _ (MatchRIds TrM p') (Ac v))) as [BJ _].
      pose proof (BI_rel_glue_states_related _ _ BJ _ HIp'm) as HBJ.
      assert (Disjoint nat (MatchRIds TrM p) (Intersection _ (MatchRIds TrM p') (Ac v))) as HDis by
        ( apply Disjoint_intro ; unfold In, not ; intros x Hx ;
          destruct Hx as [x HIx HI'Acx] ;
          destruct HI'Acx as [x HI'x HAcx] ;
          pose proof (Acv_i_st_n v x HAcx) as Hxn ;
          apply (HC1 x) ; assumption ).
      pose proof (HRuleSystem_cascading _ _ HDis sm pm p'm BI) as HCas.
      rewrite HIJ in HCas.
      destruct (HPatternNetworks_bisimilar (MatchRIds TrM p')) as [BIJ _].
      destruct (HCas BIJ HIp'm) as [HBIJ _] ; clear HCas.
      exists (Cmps sm p'm (Intersection state_type (MatchRIds TrM p') (Ac v))),
             (Cmps pm p'm (Intersection state_type (MatchRIds TrM p') (Ac v))), BIJ.
      destruct HMp'm as [_ [_ Hmp'm]].
      split.
        (* Composed s'm maps on s' *)
        destruct HMsm as [_ [_ Hmsm]].
        split.
          assert (Trans (SystemLTS M) a s s') as HTM2 by (exists v ; split ; assumption).
          destruct (trans_cond _ _ _ _ HTM2) as [_ [HSs' _]] ; exact HSs'.
        split ; try ( destruct (BI_rel_states _ _ BIJ _ _ (HBIJ HBI)) as [HS _] ; exact HS ).
        intros j Hjn HMj ; destruct (Hs' j Hjn) as [Hs'S Hs'N].
        destruct (Cmps_In sm p'm (Intersection state_type (MatchRIds TrM p') (Ac v)) j Hjn) as [Hs'm Hsm].
        destruct (EnsembleNat_dec (Ac v) j) as [Hv | Hv].
          (* Ac v j *)  unfold nthS ; rewrite Hs'm ;
            try (specialize HIp'm2 with j ; rewrite <- Init_L_to_Init_R in HIp'm2 ;
                   pose proof (mR_init_cond _ (TrG TrM j) _ (HIp'm2 HMj Hv)) as HX ;
                   unfold nthS in HX ; rewrite HX ) ;
            [ rewrite (Hs'S Hv) ; apply (Hmp'm j Hjn) | ] ; split ; assumption.
          (* ~ Ac v j*) rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HTM j Hjn Hv).
          assert (~Intersection state_type (MatchRIds TrM p') (Ac v) j) as HI'j by
            ( apply Intersection_DeMorgan ; right ; exact Hv ).
          unfold nthS ; rewrite (Hsm HI'j).
          apply Hmsm ; try assumption.
          unfold MatchRIds, MatchIds in HMj |- *.
          rewrite (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn Hv) ; exact HMj.
      split.
        (* composed p'm maps on p' *)
        destruct HMpm as [_ [_ Hmpm]].
        split ; try assumption.
        split ; try ( destruct (BI_rel_states _ _ BIJ _ _ (HBIJ HBI)) as [_ HS] ; exact HS ).
        intros j Hjn HMj.
        destruct (Cmps_In pm p'm (Intersection state_type (MatchRIds TrM p') (Ac v)) j Hjn) as [Hp'm Hpm].
        destruct (EnsembleNat_dec (Ac v) j) as [Hv | Hv].
          (* Ac v j *)  unfold nthS ; rewrite Hp'm ; [ apply (Hmp'm j Hjn) | ] ; split ; assumption.
          (* ~ Ac v j*) rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn Hv).
          assert (~Intersection state_type (MatchRIds TrM p') (Ac v) j) as HI'j by
            ( apply Intersection_DeMorgan ; right ; exact Hv ).
          unfold nthS ; rewrite (Hpm HI'j).
          apply Hmpm ; try assumption.
          unfold MatchRIds, MatchIds in HMj |- *.
          rewrite (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn Hv) ; exact HMj.
        (* BJ s'm p'm *)
        exact (HBIJ HBI).

  (* Case: A substate of Ac part of s is removed *)
  apply not_all_not_exD2 in HC1.
  destruct (classic (exists i, i < n /\ Ac v i /\
     exists q q', mR _ (TrG TrM i) q (nthS i p) /\ mR _ (TrG TrM i) q' (nthS i p') /\ 
        exists ai, nth i (sync v) None = Some ai /\ Trans (R (nth i (Rules RS) d_tr)) ai q q')) as [HC2 | HC2].

    (* Case: There is a matching transition *)
    destruct HC2 as [i [Hin [HAci HMTrans]]].
    assert (MatchRIds TrM p i) as HMpi.
      destruct HMTrans as [q [_ [Hmq _]]].
      split ; try exact Hin.
      exists q ; exact Hmq
    (* End assert M(s) i *).
    destruct (HC i Hin) as [_ HCM].
    destruct (HCM HMpi) as [sm [pm [B [HMsm [HMpm HB]]]]] ; clear HCM.
    destruct (Ac_MappedR_prop TrM HSyncCond2 HSyncCond4 v p p' pm i HT HMpm HMTrans) as [p'm [HMp'm HTR]].
    assert (Trans (SystemLTS (RightNetwork RS (MatchRIds TrM p))) a pm p'm)
      as HTR2 by ( simpl ; exists v ; split ; try exact Hva ; exact HTR ).
    destruct (BI_rel_branching_sym _ _ B _ _ HB _ _ HTR2) as [HTB | HTB].
      (* Case: a = tau /\ B sm p'm *)
      left.
      destruct HTB as [Ha HB2].
      split ; try exact Ha.
      (* C s p' *) exact (C_relate_TransR v p p' pm p'm s sm B HMpm HMp'm HMsm HT HTR HCIds HCsp HB2).
      (* Case: pm ==> phm -a-> p'm *)
      right.
      destruct HTB as [shm [s'm [HTPL [HTL [HBh HB']]]]].
      destruct (MapL_TauPath TrM (MatchRIds TrM p) _ _ HTPL _ HMsm) as [sh [HMshm [HTP2 Heqsh]]].
      destruct HTL as [v' [Hv' HTL]].
      destruct (MapL_trans TrM (MatchRIds TrM p) _ _ _ _ HTL HMshm) as [s' [HMs'm [HT2 Heqs']]].
      exists sh, s'.
      split ; try exact HTP2.
      split ; try (exists v' ; split ; assumption).
      assert (C sh p) as HCshp.
        unfold C, C_rel, Bhat_rel.
        rewrite <- (TauPath_MatchL_prop _ _ _ _ _ _ HMsm HMshm Heqsh HTPL).
        rewrite HCIds, (Union_same (MatchRIds TrM p)).
        intros j Hjn.
        split ; intro HMpj.
        split ; [rewrite <- (Heqsh j Hjn HMpj) ; apply HSs ; exact Hjn | ].
        rewrite <- (Heqsh j Hjn HMpj).
        apply HCsp ; assumption.
        exists shm, pm, B ; intuition
        (* End assert C sh p *).
      split ; try exact HCshp.
      (* C s' p' *) 
      rewrite (TauPath_MatchL_prop _ _ _ _ _ _ HMsm HMshm Heqsh HTPL),
              (Trans_MatchL_prop _ _ _ _ _ _ _ HMshm HMs'm HT2 HTL) in HCIds.
      apply (C_relate_TransR v p p' pm p'm s' s'm B HMpm HMp'm HMs'm HT HTR HCIds) ; intuition.
      rewrite <- Heqs', <- Heqsh ; try apply HCsp ; try assumption.

    (* Case: There is no matching transition *)
    right.
    assert (forall i, i < n -> Ac v i -> 
      ~(exists q q', mR _ (TrG TrM i) q (nthS i p) /\ mR _ (TrG TrM i) q' (nthS i p') 
           /\ exists ai, nth i (sync v) None = Some ai /\ Trans (R (nth i (Rules RS) d_tr)) ai q q')) 
      as HC3 by ( apply not_ex_all_notD ; apply HC2 ) ; clear HC2.
    destruct HC1 as [i [Hin [HAci HMpi]]].
    pose proof HAci as [ai Hai].
    destruct (HC i Hin) as [_ HC'].
    destruct (HC' HMpi) as [sm [pm [B [HMsm [HMpm HB]]]]] ; clear HC'.
    pose proof HMpm as [_ [HSpm _]].
    assert (forall j, j < n -> MatchRIds TrM p j ->
              Ac v j -> Init (L (nth j (Rules RS) d_tr)) (nthS j pm)) as HIAcpm.
      intros j Hjn HMpj HAcj.
      pose proof HAcj as [aj Haj].
      destruct HMpm as [_ [_ Hmpm]].
      specialize Hmpm with j.
      rewrite Init_L_to_Init_R.
      apply (TransPartlyMatchedR_prop2 TrM _ p p' (nthS j pm) j HT HAcj) ;
      try (left ; apply Hmpm) ; assumption
    (* End assert Init *).
    destruct (TauPath_glue_state_R_prop RSk (MatchRIds TrM p) (Ac v) B sm pm HB HIAcpm) 
      as [shm [HTPL [HBshm Heqshm]]].
    destruct (MapL_TauPath TrM (MatchRIds TrM p) sm shm HTPL s HMsm) as [sh [HMshm [HTP Heqsh]]].
    assert (forall j, j < n -> MatchRIds TrM p j -> Ac v j -> nthS j sh = nthS j p) as Heqsh2.
      intros j Hjn HIj HAcj.
      apply (m_injective _ _ (mR _ (TrG TrM j)) _ _ (nthS j pm)).
      destruct HMpm as [_ [_ Hmpm]].
      destruct HMshm as [_ [_ Hmshm]].
      left ; split.
      rewrite <- (mR_init_cond _ (TrG TrM j) _ (HIAcpm j Hjn HIj HAcj)).
      unfold nthS.
      rewrite (Heqshm j Hjn HIj HAcj).
      apply Hmshm ; assumption.
      apply Hmpm ; assumption
    (* End assert *).
    pose proof (Cmps_In sh p' (Ac v)) as Hs'.
    remember (Cmps sh p' (Ac v)) as s'.
    assert (NetTransSL M v sh s') as HTM.
      pose proof HT as HT' ; destruct HT as [Hv HT].
      split.
      (* SyncN M v *)
      rewrite SyncN_cond in Hv.
      apply (unchanged_transitions_v_in_SynN_M TrM HSyncCond2 HSyncCond5 v _ _ HT').
      exists i ; split ; try exact HAci.
      apply HC3 ; assumption
      (* End SyncN M v *).
      intros j Hjn.
      clear HT' ; pose proof (HT j Hjn) as HT'.
      rewrite (Procs_cond TrM) in HT' ; try exact Hjn.
      destruct (Hs' j Hjn) as [Hs'S Hs'N].
      case_eq (nth j (sync v) None) ; intros ; rewrite H in HT' ; unfold nthS.
        (* v[j] = a0 *)
        apply (TT_cond _ (TrG TrM j) a0 (nthS j p) (nthS j p')) in HT'.
        unfold nthS in Heqsh2, Heqsh.
        pose proof (AcSome_prop _ _ _ H) as HAcj.
        rewrite (Hs'S HAcj).
        destruct HT' as [[HT' _] | HT'].
        destruct (EnsembleNat_dec (MatchRIds TrM p) j) as [HMpj | HNMpj] ;
          [ (* j in M(p) *)
            rewrite (Heqsh2 j Hjn HMpj HAcj)
          | (* j not in M(p) *)
            rewrite <- (Heqsh j Hjn HNMpj) ;
            destruct (HC j Hjn) as [HC' _] ;
            unfold nthS in HC' ;
            destruct (HC' HNMpj) as [_ HRW] ;
            rewrite HRW ] ;
          exact HT'.
          exfalso ; apply (HC3 j Hjn HAcj).
          destruct HT' as [q [q' HT']].
          exists q, q' ; intuition.
          exists a0 ; intuition.
        (* v[j] = * *)
        apply AcNone_prop in H.
        rewrite (Hs'N H).
        destruct HMshm as [HSsh _].
        simpl in HSsh.
        unfold NetStates in HSsh.
        specialize HSsh with j.
        intuition
    (* End assert Trans *).
    assert (Trans (SystemLTS M) a sh s') as HTM2 by (exists v ; split ; assumption).
    exists sh, s'.
    split ; try exact HTP.
    split ; try (simpl ; exists v ; split ; assumption).
    assert (C sh p) as HCshp 
      by (apply (C_related_TauPathL p s sh B HC_rel Heqsh) ; exists pm, sm, shm ; intuition).
    split.
      (* C s ph *) exact HCshp.
      (* C s' p' *)
      unfold C_rel, D_rel, Bhat_rel ; clear Hin HAci HMpi Hai ai i.
      assert (forall i, Ac v i -> forall q', (mR _ (TrG TrM i)) q' (nthS i p') -> 
                Init (R (nth i (Rules RS) d_tr)) q') as HIRAc'.
        intros i Hv q' Hmq'.
        pose proof (Acv_i_st_n v i Hv) as Hin.
        pose proof Hv as [a' Hva'].
        apply (TransPartlyMatchedR_prop TrM v a' _ _ q' i HT Hva') ; right.
        split ; try exact Hmq'.
        intro q ; apply not_and_or.
        unfold not in HC3 |- * ; intro HTm.
        apply HC3 with (i := i) ; try assumption.
        exists q, q' ; intuition.
        exists a' ; intuition
        (* End assert Init *).
      pose proof (ComposedToState_MatchRIds TrM v p p' sh s' HT Heqs' 
                    (C_related_NetMatchIds_i_prop sh p HCshp) HIRAc') as HCIdp'.
      rewrite HCIdp', (Union_same (MatchRIds TrM p')).
      intros i Hin.
      split ; [intro HMp'i | intro HMp'i].
        (* i is not in M(p) *)
        specialize HC with i.
        destruct (Hs' i) as [Hs'Some Hs'None] ; [exact Hin | ].
        destruct (EnsembleNat_dec (Ac v) i) as [Hv | Hv].
          (* i in Ac(v) *)
          unfold nthS ; rewrite (Hs'Some Hv).
          pose proof (HSp' i Hin) as HSs'i.
          rewrite Procs_cond in HSs'i ; try exact Hin.
          rewrite (ST_cond _ (TrG TrM i)) in HSs'i.
          apply Union_to_or in HSs'i.
          destruct HSs'i as [HSs'i | HSs'i].
            (* p' was not added *) destruct HSs'i ; split ; try assumption ; try reflexivity.
            (* p' was added *)     exfalso ; apply HMp'i ; split ; assumption.
          (* i not in Ac(v) *)
          rewrite <- (NotAcMatchR_prop TrM v _ _ HT i Hin Hv) in HMp'i.
          rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT i Hin Hv).
          unfold nthS in Heqsh |- * ; rewrite (Hs'None Hv).
          rewrite <- (Heqsh i Hin HMp'i).
          destruct HC as [HC _] ; [exact Hin | ].
          apply (HC HMp'i).
        (* i is in M(s') *)
        
        
        
        set (I := Intersection nat (MatchRIds TrM p) (Ac v)).
        set (J := Setminus nat (MatchRIds TrM p') (Ac v)).
        assert (Disjoint nat J I) as HDis_J_I by (unfold I, J ; apply Disjoint_Intersection_Setminus).
        assert (Setminus nat (MatchRIds TrM p) (Ac v) = Setminus nat (MatchRIds TrM p') (Ac v)) as HJ.
          apply Extensionality_Ensembles.
          unfold Same_set, Included, In ; split ; intros x Hx ;
          destruct Hx as [HMpmx HAcx] ; pose proof HMpmx as [Hxn _] ;
          split ; unfold In ;
          solve [ rewrite <- (NotAcMatchR_prop TrM _ _ _ HT x Hxn HAcx) ; exact HMpmx
                | rewrite (NotAcMatchR_prop TrM _ _ _ HT x Hxn HAcx) ; exact HMpmx
                | exact HAcx ]
          (* End assert Setminus... *).
        pose proof (Cmps_In (d_staten n) pm J) as HpmJ.
        remember (Cmps (d_staten n) pm J) as pmJ.
        pose proof (Cmps_In (d_staten n) shm J) as HshmJ.
        remember (Cmps (d_staten n) shm J) as shmJ.
        pose proof (Cmps_In (d_staten n) pm I) as HpmI.
        remember (Cmps (d_staten n) pm I) as pmI.
        assert (NetInit (LeftNetwork RS I) pmI) as HIpmI.
          apply LeftNetwork_Init_lift.
          unfold nthS.
          intros x Hxn ; split ; intro HIx ; 
            [destruct (HpmI x Hxn) as [HpmI2 _] | destruct (HpmI x Hxn) as [_ HpmI2] ] ; rewrite (HpmI2 HIx).
          destruct HIx ; apply HIAcpm ; assumption.
          unfold d_state ; rewrite d_staten_nth ; reflexivity
          (* End assert NetInit *).
        destruct (HPatternNetworks_bisimilar J) as [BJ _].
        pose proof (HRuleSystem_cascading J I HDis_J_I shmJ pmJ pmI BJ) as HCas.
        assert ((MatchRIds TrM p) = (Union nat J I)) as HU.
          apply Extensionality_Ensembles.
          unfold I, J, Same_set, Included, In.
          rewrite <- HJ.
          split ; intros x Hx.
          (* x in M(s) *)
          destruct (classic ((Ac v) x)) as [HAcx | HAcx ].
            (* x in Ac *) apply Union_intror ; apply Intersection_intro ; assumption.
            (* x not in Ac *) apply Union_introl ; split ; assumption.
          (* x not in M(s) *)
          destruct Hx as [x Hx | x Hx].
            (* x in J *) destruct Hx as [HMpx HAcx] ; exact HMpx.
            (* x in I *) destruct Hx ; assumption
          (* End assert MRids s = Union J I *).
        rewrite <- HU in HCas ; clear HU.
        destruct (HCas B HIpmI) as [_ HBJ] ; clear HCas.
        rewrite HeqpmJ, HeqpmI, HeqshmJ in HBJ.
        unfold I, J in HBJ.
        assert (forall x, I x -> nth x pm dummy = nth x pm dummy) as Hreq1 by (intros ; reflexivity).
        assert (forall x, x < n -> ~ MatchRIds TrM p x -> nthS x pm = d_state) as Hreq2 
          by (intros x Hxn HMpx ; apply (RightNetwork_States_simpl _ _ _ x HSpm Hxn) ; exact HMpx).
        pose proof (Cas_Cmps_state_prop v p pm pm _ Hreq1 Hreq2) as Hrw ; rewrite HJ in Hrw.
        rewrite Hrw in HBJ ; clear Hreq1 Hreq2 Hrw.
        assert (forall x, I x -> nth x pm dummy = nth x shm dummy) as Hreq1 
          by ( intros x HIx ; destruct HIx as [x HMpmx HAcx] ; pose proof HMpmx as [Hxn _] ;
               rewrite (Heqshm x Hxn HMpmx HAcx) ; reflexivity ).
        assert (forall x, x < n -> ~ MatchRIds TrM p x -> nthS x shm = d_state) as Hreq2
          by ( intros x Hxn HMpx ; destruct (B_states _ _ _ _ _ HBshm) as [HSshm _] ;
               destruct (LeftNetwork_States_simpl _ _ _ x HSshm Hxn) as [_ Hshmx] ; apply Hshmx ; assumption ).
        pose proof (Cas_Cmps_state_prop v p pm shm _ Hreq1 Hreq2) as Hrw ; rewrite HJ in Hrw.
        rewrite Hrw in HBJ ; clear Hreq1 Hreq2 Hrw.
        unfold I, J in * ; rewrite <- HeqshmJ in HBJ ; rewrite <- HeqpmJ in HBJ.
        clear HIpmI HpmI HeqpmI pmI HDis_J_I I.
        set (I := Intersection nat (MatchRIds TrM p') (Ac v)).
        destruct (exists_mapRState TrM I p' HSp') as [p'mI [HMp'mI HIN]].
          intros j HIj ; destruct HIj as [j HMp'j _] ; exact HMp'j.
        assert (NetInit (LeftNetwork RS I) p'mI) as HIp'mI.
          apply LeftNetwork_Init_lift.
          intros j Hjn.
          rewrite Init_L_to_Init_R.
          split ; intro HIj.
          (* j in I *)
          destruct HIj as [j HMp'j HAcj].
          destruct HMp'mI as [_ [_ Hmp'mI]].
          apply HIRAc' ; [exact HAcj | apply Hmp'mI ; try exact Hjn ; split ; assumption].
          (* j not in I *)
          apply DummyLTS_Init ; apply HIN ; try unfold I ; assumption
        (* End assert *).
        destruct (HPatternNetworks_bisimilar (Union _ J I)) as [BIJ _].
        assert (Disjoint nat J I) as HDis_J_I by (unfold I, J ; apply Disjoint_Intersection_Setminus).
        destruct (HRuleSystem_cascading J I HDis_J_I shmJ pmJ p'mI BJ BIJ HIp'mI) as [HCas _].
        pose proof (HCas (HBJ HBshm)) as HBIJ.
        rewrite <- (Union_Setminus_Intersection (MatchRIds TrM p') (Ac v)).
        exists (Cmps shmJ p'mI I), (Cmps pmJ p'mI I), BIJ.
        split.
          (* shmJ[p'mJ / J] maps on s' *)
          split.
          destruct (trans_cond _ _ _ _ HTM2) as [_ [HSs' _]].
          exact HSs'.
          split.
          destruct (BI_rel_states RSk _ _ _ _ HBIJ) as [HSs'm _].
          apply HSs'm.
          intros j Hjn HIJj.
          destruct (Hs' j Hjn) as [Hs'S Hs'N].
          destruct HIJj as [j HJj | j HIj].
            (* j in J *)
            unfold nthS.
            pose proof (Disjoint_In _ _ HDis_J_I j HJj) as HNIj.
            rewrite (Cmps_i_not_in_I _ p'mI I j Hjn HNIj).
            destruct (HshmJ j Hjn) as [HshmJ' _].
            rewrite (HshmJ' HJj).
            destruct HJj as [HMp'j HAcj] ; unfold In in HMp'j.
            rewrite (Hs'N HAcj).
            destruct HMshm as [_ [_ Hmshm]].
            rewrite <- (NotAcMatchR_prop TrM _ _ _ HT j Hjn HAcj) in HMp'j.
            apply Hmshm ; assumption.
            (* j in I *)
            unfold nthS.
            rewrite (Cmps_i_in_I shmJ p'mI I j Hjn HIj).
            destruct (HshmJ j Hjn) as [HshmJ' _].
            destruct HMp'mI as [_ [_ Hmp'mI]].
            destruct (LeftNetwork_Init_simpl RS I p'mI j HIp'mI Hjn) as [HIp'mI2 _].
            pose proof (mR_init_cond _ (TrG TrM j) _ (HIp'mI2 HIj)) as HRW ; unfold nthS in HRW.
            rewrite HRW ; clear HRW.
            assert (Ac v j) as HAcj by (destruct HIj ; assumption).
            rewrite (Hs'S HAcj).
            apply Hmp'mI ; assumption. 
        split.
          (* p'mI[pmJ / J] maps on p' *)
          split ; try exact HSp'.
          split.
          destruct (BI_rel_states RSk _ _ _ _ HBIJ) as [_ HSp'm].
          apply HSp'm.
          intros j Hjn HIJj.
(*          destruct (Hp'm j Hjn) as [Hp'mJ Hp'mN]. *)
          destruct HIJj as [j HJj | j HIj].
            (* j in J *)
            pose proof HJj as [HMp'j HNAcj] ; unfold In in HMp'j.
            rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT j Hjn HNAcj).
            rewrite <- (NotAcMatchR_prop TrM _ _ _ HT j Hjn HNAcj) in HMp'j.
            unfold nthS.
            pose proof (Disjoint_In _ _ HDis_J_I j HJj) as HNIj.
            rewrite (Cmps_i_not_in_I _ _ I j Hjn HNIj).
            destruct (HpmJ j Hjn) as [HpmJ1 _].
            rewrite (HpmJ1 HJj).
            destruct HMpm as [_ [_ Hmpm]].
            apply Hmpm ; assumption.
            (* j in I *)
            unfold nthS.
            rewrite (Cmps_i_in_I _ _ I j Hjn HIj).
            destruct HMp'mI as [_ [_ Hmp'mI]].
            apply Hmp'mI ; assumption.
         (* s' B p' *)
         assumption.
Qed.

Proposition Check_Correct :
  cond1_universal_applicabilityL TrM -> 
  cond2_universal_applicabilityR TrM -> 
  cond3_completenessL RS -> 
  cond4_completenessR RS ->
  cond5_new_SyncLaw M RS ->
  RuleSystem_cascading RSk ->
  PatternNetworks_bisimilar RSk ->
    exists (C : BB (SystemLTS M) (SystemLTS TrM)), True.
Proof.
intros HSyncCond1 HSyncCond2 HSyncCond3 HSyncCond4 HSyncCond5 HCascading HPatternsBisimilar.
exists 
  (Build_BB n (SystemLTS M) (SystemLTS TrM) C_rel
    (C_rel_states)
    (C_rel_init HPatternsBisimilar)
    (C_rel_branching HSyncCond1 HSyncCond3 HCascading HPatternsBisimilar)
    (C_rel_branching_sym HSyncCond2 HSyncCond4 HSyncCond5 HCascading HPatternsBisimilar)).
trivial.
Qed.

End proposition_proof.

End NetworkTransformationCheck.