Module ClassicalExt.

(* ------------------------------------------------
     Classical logic
   ------------------------------------------------ *)
Require Export Classical.

Lemma Option_NNone_Some {A} (o : option A) : o <> None -> exists a : A, o = Some a.
Proof.
intros.
destruct o.
exists a ; reflexivity.
congruence.
Qed.

Lemma Option_Some_NNone {A} (o : option A)(a : A) : o = Some a -> o <> None.
Proof.
intros.
destruct o.
unfold not ; intro ; congruence.
congruence.
Qed.

Lemma Option_NSome_None {A} (o : option A) : ~(exists a : A, o = Some a) -> o = None.
Proof.
intros.
destruct o.
apply NNPP.
unfold not in * ; intros.
apply H.
exists a.
reflexivity.
reflexivity.
Qed.

Lemma contrapositive_to_not : forall (P Q : Prop),(P -> Q) -> (~Q -> ~P).
Proof.
intuition.
Qed.

Lemma contrapositive_from_not : forall (P Q : Prop), (~P -> ~Q) -> (Q -> P).
Proof.
intuition.
apply NNPP.
unfold not ; intro.
apply H.
exact H1.
exact H0.
Qed.

Lemma contrapositive : forall (P Q : Prop), (P -> Q) <-> (~Q -> ~P).
Proof.
intros ; split.
apply contrapositive_to_not.
apply contrapositive_from_not.
Qed.

Lemma contrapositive_to_not_1 : forall {A} (P Q : A -> Prop), 
  (forall a : A, P a -> Q a) -> (forall a : A, ~Q a -> ~P a).
Proof.
intuition.
Qed.

Lemma contrapositive_from_not_1 : forall {A} (P Q : A -> Prop), 
  (forall a : A, ~P a -> ~Q a) -> (forall a : A, Q a -> P a).
Proof.
intros A P Q H a.
specialize H with a.
apply contrapositive_from_not.
exact H.
Qed.

Lemma not_ex_all_notD : forall A (D P Q: A -> Prop),
  (forall a, D a -> P a -> ~Q a) <-> ~(exists a, D a /\ P a /\ Q a).
Proof.
intuition.
destruct H0 as [a [HDa [HPa HQa]]].
apply (H a HDa HPa).
exact HQa.
apply H.
exists a.
split ; try split ; try assumption.
Qed.

Lemma not_all_not_exD1 : forall A (D1 P : A -> Prop),
  ~(forall a, D1 a -> ~P a) <-> exists a, D1 a /\ P a.
Proof.
intuition.
apply NNPP.
intuition.
apply H.
intros.
apply H0.
exists a.
intuition.
destruct H as [a Ha].
apply (H0 a) ; intuition.
Qed.

Lemma not_all_not_exD2 : forall A (D1 D2 P : A -> Prop),
  ~(forall a, D1 a -> D2 a -> ~P a) <-> exists a, D1 a /\ D2 a /\ P a.
Proof.
intuition.
apply NNPP.
intuition.
apply H.
intros.
apply H0.
exists a.
intuition.
destruct H as [a Ha].
apply (H0 a) ; intuition.
Qed.

Lemma forall_not_or_2_not_exists_and : forall A (P Q : A -> Prop), 
  (forall a, ~P a \/ ~Q a) <-> ~(exists a, P a /\ Q a).
Proof.
intuition.
destruct H0 as [a Ha].
specialize H with a.
destruct H ;
apply H ; intuition.
apply not_and_or.
unfold not ; intro.
apply H.
exists a ; exact H0.
Qed.

End ClassicalExt.