Module ListN.

Load "Utils.ListProps".
Export ListProps.

(* ------------------------------------------------
     listn (a vector)
   ------------------------------------------------ *)

(* a list of length n *)
Record listn (A : Type)(n : nat) := {
  lst :> list A;
  lst_length : length lst = n
}.

Definition Bld_listn {A}{n}(l : list A)(H : length l = n) := Build_listn A n l H.
Definition Bld_list1 {A}(a : A) := Build_listn A 1 (a :: nil) (list1_length a).

Lemma listn_eq :
  forall {A}{n} (l1 l2 : list A)(H1 : length l1 = n)(H2 : length l2 = n), 
    l1 = l2 -> Build_listn A n l1 H1 = Build_listn A n l2 H2.
Proof.
intros.
destruct H.
f_equal.
apply proof_irrelevance.
Qed.

Lemma lstn_eq :  forall {A}{n} (l1 l2 : @listn A n)(d : A), 
                   (forall i, i < n -> nth i l1 d = nth i l2 d) -> l1 = l2.
Proof.
intros.
destruct l1, l2.
apply listn_eq.
apply (list_eq lst0 lst1 d).
intuition.
intros.
specialize H with i.
simpl in H.
assert (i < n \/ i >= n) by omega.
destruct H0.
intuition.
assert (length lst0 <= i) by omega.
assert (length lst1 <= i) by omega.
rewrite (nth_overflow _ d H1).
rewrite (nth_overflow _ d H2).
reflexivity.
Qed.

Lemma listn_listn_eq : forall {A}{n} (l1 l2 : listn A n), lst A n l1 = lst A n l2 -> l1 = l2.
Proof.
intros.
destruct l1, l2.
apply listn_eq.
auto.
Qed.

Lemma listn_lst_eq : 
  forall {A}{n} (l1 l2 : list A)(H1 : length l1 = n)(H2 : length l2 = n),
    Build_listn A n l1 H1 = Build_listn A n l2 H2 -> l1 = l2.
Proof.
intros.
congruence.
Qed.

Lemma list1_eq : 
  forall {A} (e1 e2 : A),
    e1 = e2 -> Bld_list1 e1 = Bld_list1 e2.
Proof.
intros.
apply listn_eq.
f_equal.
exact H.
Qed.

Lemma list1_lst_eq : forall {A} (e1 e2 : A), Bld_list1 e1 = Bld_list1 e2 -> e1 = e2.
Proof.
intros.
unfold Bld_list1 in H.
congruence.
Qed.

Lemma list1_prop : forall {A}(s : listn A 1)(d : A), Bld_list1 (hd d s) = s.
Proof.
intros.
pose proof (listn_listn_eq s (Bld_list1 (hd d s))).
unfold Bld_list1 in *.
destruct s.
simpl in *.
pose proof (vec1_prop d lst0).
intuition.
Qed.

Lemma listn_nth_bound : forall {A}{n} (d : A)(l : listn A n) i, nth i l d <> d -> i < n.
Proof.
intros.
rewrite <- (lst_length A n l).
apply (list_nth_bound d l i).
exact H.
Qed.

Lemma listn_forall_exists : 
  forall A n (d : A) (P : nat -> A -> Prop), 
    (forall i, i < n -> exists (a : A), P i a) -> 
      (exists (l : listn A n), forall i, i < n -> P i (nth i l d)).
Proof.
intros.
induction n.
(* Base case *)
assert (length (@nil A) = 0) by intuition.
exists (Build_listn A 0 nil H0).
intros.
omega.
(* Step case *)
destruct IHn as [l Hl] ; intuition.
specialize H with n.
destruct H as [a Ha] ; auto.
assert (length (l ++ (a :: nil)) = S n).
  rewrite app_length.
  simpl.
  assert (forall x, x + 1 = S x) by intuition.
  rewrite H.
  f_equal.
  apply lst_length (* End assert *).
exists (Build_listn A (S n) (l ++ (a :: nil)) H).
intros.
assert (i < n \/ i = n) by omega.
destruct H1.
(* i < n *)
simpl.
specialize Hl with i.
assert (i < length l).
pose proof (lst_length A n l).
rewrite H2.
exact H1 (* End assert *).
rewrite (app_nth1 l (a :: nil) d H2).
apply Hl.
exact H1.
(* i = n *)
rewrite H1.
simpl.
assert (n >= length l).
pose proof (lst_length A n l).
omega.
rewrite (app_nth2 l (a :: nil) d H2).
rewrite (lst_length A n l).
assert (n - n = 0) by omega.
rewrite H3.
simpl.
exact Ha.
Qed.

Lemma listn_forall_exists2 : 
  forall A n (d : A) (P : nat -> A -> A -> Prop), 
    (forall i, i < n -> exists (a1 a2 : A), P i a1 a2) -> 
      (exists (l1 l2 : listn A n), forall i, i < n -> P i (nth i l1 d) (nth i l2 d)).
Proof.
intros.
induction n.
(* Base case *)
assert (length (@nil A) = 0) by intuition.
exists (Build_listn A 0 nil H0), (Build_listn A 0 nil H0).
intros.
omega.
(* Step case *)
destruct IHn as [l1 [l2 Hl]] ; intuition.
specialize H with n.
destruct H as [a1 [a2 Ha]] ; auto.
assert (forall (a : A) l, length (l ++ (a :: nil)) = S (length l)).
  intros.
  rewrite app_length.
  simpl.
  assert (forall x, x + 1 = S x) by intuition.
  rewrite H.
  f_equal (* End assert *).
pose proof (H a1 l1).
pose proof (H a2 l2).
rewrite lst_length in H0.
rewrite lst_length in H1.
exists (Build_listn A (S n) (l1 ++ (a1 :: nil)) H0),
       (Build_listn A (S n) (l2 ++ (a2 :: nil)) H1).
intros.
assert (i < n \/ i = n) by omega.
destruct H3.
(* i < n *)
simpl.
specialize Hl with i.
assert (forall l : listn A n, i < length l).
  intros.
  pose proof (lst_length A n l).
  rewrite H4.
  exact H3 (* End assert *).
rewrite (app_nth1 l1 (a1 :: nil) d (H4 l1)).
rewrite (app_nth1 l2 (a2 :: nil) d (H4 l2)).
apply Hl.
exact H3.
(* i = n *)
rewrite H3.
simpl.
assert (forall l : listn A n, n >= length l).
  intros.
  pose proof (lst_length A n l).
  omega (* End assert *).
rewrite (app_nth2 l1 (a1 :: nil) d (H4 l1)).
rewrite (app_nth2 l2 (a2 :: nil) d (H4 l2)).
rewrite (lst_length A n l1).
rewrite (lst_length A n l2).
assert (n - n = 0) by omega.
rewrite H5.
simpl.
exact Ha.
Qed.

End ListN.