Module ListProps.

(* ------------------------------------------------
     List properties
   ------------------------------------------------ *)

Require Export Program.Equality.
Require Export List.
Require Import Omega.

Lemma list_nth_bound : forall {A} (d : A)(v : list A) i, nth i v d <> d -> i < length v.
Proof.
intros.
dependent induction v.
exfalso.
apply H.
destruct i ;
simpl ; reflexivity.
destruct i.
simpl in *.
omega.
simpl in *.
specialize IHv with i.
intuition.
Qed.

Lemma list_prop : forall {A} (d : A)(v : list A), length v > 0 -> v = (hd d v) :: (tl v).
Proof.
intros.
induction v.
simpl in H.
intuition.
simpl in *.
reflexivity.
Qed.

Lemma vec0_prop : forall {A} (v : list A), length v = 0 -> v = nil.
Proof.
intros.
induction v.
reflexivity.
contradict H.
simpl.
congruence.
Qed.

Lemma vec1_prop : forall {A} (d : A)(v : list A), length v = 1 -> v = (hd d v) :: nil.
Proof.
intros.
assert (tl v = nil).
dependent destruction v.
simpl.
reflexivity.
simpl in H.
simpl.
assert (length v = 0).
intuition.
apply vec0_prop in H0.
exact H0.
pose proof (list_prop d v).
rewrite H0 in H1.
apply H1.
omega.
Qed.

Lemma vec1_eq : forall {A} (d : A)(v1 v2 : list A), 
  length v1 = 1 /\ length v2 = 1 -> hd d v1 = hd d v2 -> v1 = v2.
Proof.
intros.
destruct H.
rewrite (vec1_prop d v1).
rewrite (vec1_prop d v2).
rewrite H0.
trivial.
exact H1.
exact H.
Qed.

Lemma list1_length : forall {A}(a : A), length (a :: nil) = 1.
Proof.
intuition.
Qed.

Lemma list_eq : forall {A} (l1 l2 : list A)(d : A), length l1 = length l2 ->
                  (forall i, nth i l1 d = nth i l2 d) -> l1 = l2.
Proof.
intros.
dependent induction l1 ;
induction l2.
reflexivity.
simpl in H.
omega.
simpl in H.
omega.
simpl in H.
assert (length l1 = length l2) by omega.
assert (forall j, nth j l1 d = nth j l2 d).
intros.
specialize H0 with (S j).
simpl in H0.
exact H0.
pose proof (IHl1 l2 d H1 H2).
specialize H0 with 0.
simpl in H0.
congruence.
Qed.

(* Build a list of a's of length n *)
Fixpoint alist {A}(n : nat)(a : A) :=
  match n with
  | S m => a :: alist m a
  | 0 => nil
  end.

Lemma alist_nth : forall {A}(n i : nat)(a : A), nth i (alist n a) a = a.
Proof.
intros.
dependent induction n.
simpl.
destruct i ;
reflexivity.
simpl.
destruct i.
reflexivity.
apply IHn.
Qed.

Lemma alist_length : forall {A}(n : nat)(a : A), length (alist n a) = n.
Proof.
intros.
induction n.
simpl.
reflexivity.
simpl.
apply f_equal.
exact IHn.
Qed.

End ListProps.