Module SetTheoryProps.

(* ------------------------------------------------
     Set Lemmas
   ------------------------------------------------ *)

Load "Utils.ClassicalExt".
Import ClassicalExt.
Require Export Constructive_sets (* Sets *).

Lemma Union_to_or : forall {A} B C (a : A), Union A B C a -> B a \/ C a.
Proof.
intros.
destruct H ; intuition.
Qed.

Lemma Union_same : forall {A} B, Union A B B = B.
Proof.
intros.
apply Extensionality_Ensembles.
unfold Same_set, Included.
split ; intros.
destruct H ; exact H.
apply Union_introl ; exact H.
Qed.

Lemma Intersection_DeMorgan : forall {A} B C (a : A), ~Intersection A B C a <-> (~B a \/ ~C a).
Proof.
intros ; split ; intros.
apply not_and_or.
unfold not in * ; intros ; apply H.
destruct H0 ; split ; assumption.
unfold not ; intros ; destruct H, H0 ; apply H ; [exact H0 | exact H1].
Qed.

Lemma Included_Intersection : forall {A} B C, Included A B C -> Intersection A B C = B.
Proof.
intros.
apply Extensionality_Ensembles.
unfold Same_set, Included in *.
split ; intros.
destruct H0.
exact H0.
split ; try apply H ; exact H0.
Qed.

Lemma Union_Setminus_Intersection : forall {A} (B C : Ensemble A),
  Union A (Setminus A B C) (Intersection A B C) = B.
Proof.
intros.
apply Extensionality_Ensembles.
unfold Same_set, Included, Setminus.
split ; intros.
destruct H.
destruct H.
exact H.
destruct H.
exact H.
destruct (classic (C x)) ; [right | left] ;
split ; assumption.
Qed.

Lemma Disjoint_Intersection_Setminus :
  forall {A} (B C D : Ensemble A),
    Disjoint A (Setminus A B C) (Intersection A D C).
Proof.
intros.
apply Disjoint_intro.
unfold not, In.
intros.
destruct H.
destruct H.
destruct H0.
contradiction.
Qed.

Lemma Disjoint_In : forall {A} (B C: Ensemble A), Disjoint A B C -> forall x, B x -> ~C x.
Proof.
unfold not.
intros.
destruct H.
specialize H with x.
apply H.
split ; assumption.
Qed.

Lemma Disjoint_In2 : forall {A} (B C: Ensemble A), Disjoint A B C -> forall x, C x -> ~B x.
Proof.
unfold not.
intros.
destruct H.
specialize H with x.
apply H.
split ; assumption.
Qed.

End SetTheoryProps.